import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.util.*;

import sv.kernel.*;
import sv.ui.*;

/**
 * CollabManager.java
 * collaboration manager
 * @version 1.0   10/1/97
 * @author Byeongseob Ki
 */

public class CollabManager {
  svserver svServer;

  private Vector conNameVec;
  private Vector conPortVec;
  private boolean isCollaboration = false;

  public ConnectionHandler conHandler;
  private ConInterface conInterface;

  public CollabManager(svserver svServer) {
     this.svServer = svServer;

     conNameVec = new Vector();
     conPortVec = new Vector();
  }

  /**
   * return conNameVec
   */
  public Vector getNameVec() {
    return conNameVec;
  }

  /**
   * return conPortVec
   */
  public Vector getPortVec() {
    return conPortVec;
  }

  /**
   * set collaboration mode
   */
  public void setCollaborationMode(boolean b) {
    isCollaboration = b;
  }


  /** 
   * action when the user click the "connect" button
   */
  public void clickedConnectButton() {
     ConInterface conInterface = new ConInterface(svServer, "Connection ");
     conInterface.show();
  }

  /** 
   * action when the user click the "exit" button
   */
  public void clickedExitButton() {
     if (isCollaboration)
       peerDisconnectRequest();
  }


  /**
   * collaboration request from slave
   * @param collaborator hostname
   * @param port collaborator host port
   */
  public void peerConnectRequest(String server, int port) {
     try {
        Socket sock = new Socket(server, port);
        DataOutputStream output = new DataOutputStream(sock.getOutputStream());

        output.writeInt(2);                     // send command
        output.writeInt(svServer.this_port);    // send port
        
     } catch (Exception e) {
        svServer.message("Connection failed ....", true);
     }
  }

  /**
   * other server requested collaboration, wait decision by showing dialog
   * @param server other server hostname
   * @param port other server host port
   */
  public void peerCollaborationRequested(String server, int port) {
     ReplyDialog dialog
            = new ReplyDialog(svServer, server, port, "Connection Requesed", 
                          "Connection requested from server " + server);
     dialog.show();
     peerInitReply(server, port, dialog.getConnectionMode(), 
                                                 dialog.isAccepted());
  }


  /**
   * reply to requested server. If accepted, broadcast new collaborator's name 
   * to all other collaborators. And then add this collaborator to connection 
   * list. Also send all initial states.
   * @param server slave hostname
   * @param port slave host port
   * @param mode if mode is 1, then master-master connection
   * @param accepted if true, accepted
   */
  public void peerInitReply(String server, int port,
                                            int mode, boolean accepted) {

     if (accepted) {

        CollSender sender = new CollSender(this, server, port, mode);
        sender.start();
        
     } else {
        try {
           Socket sock = new Socket(server, port);
           DataOutputStream output =
                   new DataOutputStream(sock.getOutputStream());
           output.writeInt(5);               // send refuse command
        } catch (Exception e) {
           svServer.message("Connection failed ....", true);
        }
     }
  }


  /**
   * received connection is established. Receive all info from collaborator
   * @param hostname from this host
   */
  public void connectionEstablished(DataInputStream in, String hostname) {
    int dummy;
    String[] configs;
    String title;

    try {
       int mode = in.readInt(); 
 
       // receive connection list size
       if (mode ==1) {
          int connectionSz = in.readInt(); 
          for (int i=0; i<connectionSz; i++) {
              conNameVec.addElement(in.readUTF());
              conPortVec.addElement(new Integer(in.readInt()));
           }

           conNameVec.addElement(hostname);
           conPortVec.addElement(new Integer(in.readInt()));
       }

       isCollaboration = true;
       if (conHandler == null) {
           conHandler = new ConnectionHandler(svServer);
           conHandler.setSize(480, 350);
           conHandler.show();
           conHandler.updateListItem();
       }

       // receive system configuration
       dummy = in.readInt();
       configs = new String[dummy];
       for (int j=0; j<dummy; j++)          
          configs[j] =  in.readUTF();     
       svServer.configHandler.config1d.doConfig(configs);

       dummy = in.readInt();
       configs = new String[dummy];
       for (int j=0; j<dummy; j++)          
          configs[j] =  in.readUTF();     
       svServer.configHandler.config2d.doConfig(configs);

       dummy = in.readInt();
       configs = new String[dummy];
       for (int j=0; j<dummy; j++)          
          configs[j] =  in.readUTF();     
       svServer.configHandler.config3d.doConfig(configs);


       // receive all frame's data set (1d)
       int noOfFrames = in.readInt(); 
       for (int i=0; i<noOfFrames; i++) 
          receiveGFPlane(in);
                         
       // receive all frame's data set (2d)
       noOfFrames = in.readInt(); 
       for (int i=0; i<noOfFrames; i++) 
          receiveGFrame2D(in);

       // receive all frame's data set (3d)
       noOfFrames = in.readInt(); 
       for (int i=0; i<noOfFrames; i++) 
          receiveGFrame3D(in);

      } catch (Exception e) {
        svServer.message("Data receiving error(4).", true);
e.printStackTrace();
      }
  }


  /**
   * received GFrame
   */
  public void receiveGFrame(DataInputStream in) {
    try {
       int dim = in.readInt();
 
       switch (dim) {
          case DataSet.ONED:
             receiveGFPlane(in);
             break;
          case DataSet.TWOD:
             receiveGFrame2D(in);
             break;
          case DataSet.THREED:
             receiveGFrame3D(in);
             break;
          default:
             System.out.println("Not support yet ...");
       }
    } catch (Exception e) {
       System.out.println("Data receiveing error(8) ...");
    }
  }



  /**
   * add new collaborator
   */
  public void addNewCollaborator(String server, Integer port, 
                                                String shortname) {
     conNameVec.addElement(server);
     conPortVec.addElement(port);
     svServer.message("New collaborator is added from " + server + " ...",
                                                                 true);
     if (conHandler != null) {
         conHandler.updateListItem();
         conHandler.addEventInfo("<" + shortname + "> New collaborator is added.");
     }
  }

  /**
   * broadcast disconnect message to all collaborator
   */
  public void peerDisconnectRequest() {
     if (!isCollaboration) return;   // nothing to do

     int size = conNameVec.size();
     for (int i=0; i<size; i++) {
        String server = (String)conNameVec.elementAt(i);
        int port = ((Integer)conPortVec.elementAt(i)).intValue();

        try {
           Socket sock = new Socket(server, port);
           DataOutputStream output = 
                           new DataOutputStream(sock.getOutputStream());

           output.writeInt(3);            // send exit collaboration command

       } catch (Exception e) {
          svServer.message("Connection failed ....", true);
       }

     } 

     isCollaboration = false;
     if (conHandler != null) {
         conHandler.dispose();
         conHandler = null;
     }
     conNameVec.removeAllElements();
     conPortVec.removeAllElements();
  }

  /**
   * received disconnect message from other collaborator
   * @param hostname from this host
   */
  public void receivedDisconnectRequest(String hostname, String shortname) {
      svServer.message("Disconnection requested from " + hostname, true);

      // remove that server or quit the collaboration
      // if no more coraborator

      int sz = conNameVec.size();
      for (int i=0; i<sz; i++) {
         String str = (String)conNameVec.elementAt(i);
         if (str.indexOf(hostname) >= 0) {
             conNameVec.removeElementAt(i);
             conPortVec.removeElementAt(i);
             break;
         }
       }

       sz = conNameVec.size();
       if (sz < 1) {
           isCollaboration = false;
           if (conHandler != null) {
              conHandler.dispose();
              conHandler = null;
           }
       }
            
       if (conHandler != null) {
          conHandler.updateListItem();
          conHandler.addEventInfo("<" + shortname + "> Disconnected");
       }     
  }

  /**
   * send the data set to all collaborator
   * @param dataSet DataSet to broadcast
   * @param id owner frame of this data set
   */
  public void broadcastDataSet(DataSet dataSet, long id) {
     CollSender sender = new CollSender(this, dataSet, id);
     sender.start();
  }

  /**
   * broadcast command
   * @param dim dimension of the target window
   * @param command command
   * @param str attached string
   * @param id frame id
   */
  public void broadcastCommand(GFrame gframe, int command, String[] args) {
     if (gframe != null && gframe.stealthMode)
        return;

     int size = conNameVec.size();
     for (int i=0; i<size; i++) {
        String sname = (String)conNameVec.elementAt(i);
        int sport = ((Integer)conPortVec.elementAt(i)).intValue();

        try {
           Socket sock = new Socket(sname, sport);
           DataOutputStream output = 
                        new DataOutputStream(sock.getOutputStream());

           output.writeInt(21);   // send command for broadcasting

           if (gframe == null)
              output.writeInt(-1);
           else
              output.writeInt(gframe.getRank());

           output.writeInt(command);
         
           if (args == null)
             output.writeInt(-1);
           else {
             output.writeInt(args.length);
             for (int j=0; j<args.length; j++)
                 output.writeUTF(args[j]); 
           }

           if (gframe == null)
              output.writeLong(-1);
           else
              output.writeLong(gframe.getID()); 

        } catch (Exception e) {
           svServer.message("Connection failed ....", true);
        }
     }
  }    


  /**
   * broadcast filter
   * @param id frame id
   * @param dim dimension of the target window
   * @param cname name of the filter
   * @param result class byte stream
   * @param input input string to filter
   */
  public void broadcastFilter(long id, int dim, String cname,
                         byte[] result, String input, String[] idArray) {

     int size = conNameVec.size();
     for (int i=0; i<size; i++) {
        String sname = (String)conNameVec.elementAt(i);
        int sport = ((Integer)conPortVec.elementAt(i)).intValue();

        try {
           Socket sock = new Socket(sname, sport);
           DataOutputStream output = 
                        new DataOutputStream(sock.getOutputStream());

           output.writeInt(11);   // send command for broadcasting

           output.writeLong(id);
           output.writeInt(dim);

           if (input == null || input.length() <= 0) output.writeInt(0);
           else {
             output.writeInt(1);
             output.writeUTF(input);
           }

           output.writeUTF(cname);

           output.writeInt(idArray.length);
           for (int j=0; j<idArray.length; j++)
              output.writeUTF(idArray[j]);

           output.writeInt(result.length);
           output.write(result, 0, result.length);

        } catch (Exception e) {
           svServer.message("Connection failed ....", true);
        }
     }
  }    



  /**
   * broadcast string(will used by chat only)
   * @param str attached string
   */
  public void broadcastString(String str) {

     int size = conNameVec.size();
     for (int i=0; i<size; i++) {
        String sname = (String)conNameVec.elementAt(i);
        int sport = ((Integer)conPortVec.elementAt(i)).intValue();

        try {
           Socket sock = new Socket(sname, sport);
           DataOutputStream output = 
                        new DataOutputStream(sock.getOutputStream());

           output.writeInt(9);   // send command for broadcasting
           output.writeUTF(str); 

        } catch (Exception e) {
           svServer.message("Connection failed ....", true);
        }
     }
  }    

  /**
   * broadcast remove data set command
   * @param id frame id, owner of the data set
   * @param data DataSet to be removed
   */
  public void broadcastRemoveDataSet(long id, DataSet data) {

     int size = conNameVec.size();
     for (int i=0; i<size; i++) {
        String sname = (String)conNameVec.elementAt(i);
        int sport = ((Integer)conPortVec.elementAt(i)).intValue();

        try {
           Socket sock = new Socket(sname, sport);
           DataOutputStream output = 
                        new DataOutputStream(sock.getOutputStream());

           output.writeInt(14);      // send command for broadcasting
           output.writeInt(data.getRank());       // send dimension
           output.writeLong(id); 
           output.writeUTF(data.getTitle());

        } catch (Exception e) {
           svServer.message("Connection failed ....", true);
        }
     }
  }


  private void receiveGFPlane(DataInputStream in) {
    int dummy;
    String[] configs;
    String title;
    int dataType;

    try {
      GFPlane gframe = null;
      boolean stealth = in.readBoolean();
      if (stealth) return;

      dataType = in.readInt();
      long frameID = in.readLong();
    
      dummy = in.readInt();
      configs = new String[dummy];
      for (int j=0; j<dummy; j++)          
         configs[j] =  in.readUTF();     

      CPlane cvalues = new CPlane(configs);

      ////////////////////////////////////////////////////////
      // receives whiteboard objects
      String drawingobjects = null;
      int wsz = in.readInt(); 
      if (wsz > 0) {
         byte[] readArr = new byte[wsz];
         in.readFully(readArr, 0, wsz);
         drawingobjects = new String(readArr);
      }
      ////////////////////////////////////////////////////////

      int noOfDataSet = in.readInt(); 

      for (int j=0; j<noOfDataSet; j++) {
         DSPlane dataSet = null;

         title = in.readUTF();

         int noOfTimeData = in.readInt(); 
         for (int k=0; k<noOfTimeData; k++) {
            TDPlane readed = null;

            if (dataType == DSPlane.NORMAL)
               readed = SerialMgr.parse1dStream(in);             
            else if (dataType == DSPlane.VECTOR)
               readed = SerialMgr.parsevecStream(in);     
            else if (dataType == DSPlane.CONNECTIVITY)
               readed = SerialMgr.parseconStream(in);     

            if (k == 0) {
                if (dataType == DSPlane.NORMAL)
                   dataSet = new DataSet1D(title, (TimeData1D)readed);
                else if (dataType == DSPlane.VECTOR)
                   dataSet = new DataSetVec(title, (TimeDataVec)readed);
                 else if (dataType == DSPlane.CONNECTIVITY)
                   dataSet = new DataSetCon(title, (TimeDataCon)readed);
            } else 
                dataSet.addTimeData(readed);
	 }

         if (j == 0) {
             long fid = svServer.fmanager.register(dataSet, frameID, cvalues);
             gframe = (GFPlane)svServer.fmanager.getGFrame(fid);
         } else
             gframe.addDataSet(dataSet);
      }
      gframe.createWhiteboard(drawingobjects);
    } catch (Exception e) {
      System.out.println("Receiving 1d window information error ...");
    }
  }    

  private void receiveGFrame2D(DataInputStream in) {
    int dummy;
    String[] configs;
    String title;

    try {
       GFrame2D gframe = null;
       boolean stealth = in.readBoolean();
       if (stealth) return;

       long frameID = in.readLong();
     
       dummy = in.readInt();
       configs  = new String[dummy];
       for (int j=0; j<dummy; j++)          
          configs[j] = in.readUTF();     

       Config2D cvalues = new Config2D(configs);
       String[] colorinfo = new String[2];
       colorinfo[0] = in.readUTF(); 
       colorinfo[1] = in.readUTF(); 

       int noOfDataSet = in.readInt(); 

       for (int j=0; j<noOfDataSet; j++) {
          DataSet2D dataSet = null;

          title = in.readUTF();

          int noOfTimeData = in.readInt(); 
          for (int k=0; k<noOfTimeData; k++) {
             TimeData2D readed = SerialMgr.parse2dStream(in);

             if (k == 0) 
                 dataSet = new DataSet2D(title, readed);
             else 
                 dataSet.addTimeData(readed);
	     }

          if (j == 0) {
              long fid = svServer.fmanager.register(dataSet, frameID, cvalues);
              gframe = (GFrame2D)svServer.fmanager.getGFrame(fid);
              if (!gframe.getColorInfo().equals(colorinfo) && !colorinfo.equals("NONE"))
                 gframe.setColorMap(colorinfo);
          } else
              gframe.addDataSet(dataSet);
       }
    } catch (Exception e) {
      System.out.println("Receiving 2d window information error ...");
    }
  }    


  private void receiveGFrame3D(DataInputStream in) {
    int dummy;
    String[] configs;
    String title;

    try {
       GFrame3D gframe = null;
       boolean stealth = in.readBoolean();
       if (stealth) return;

       long frameID = in.readLong();
     
       dummy = in.readInt();
       configs  = new String[dummy];
       for (int j=0; j<dummy; j++)          
          configs[j] = in.readUTF();     

       Config3D cvalues = new Config3D(configs);

       DataSet3D dataSet = null;

       title = in.readUTF();

       String t_hn = null;
       int t_port = -1;
       int t_ver = -1;

       if (in.readInt() == 1) {
          t_hn = in.readUTF();
          t_port = in.readInt();
          t_ver = in.readInt();
       }

       int noOfTimeData = in.readInt(); 
       for (int k=0; k<noOfTimeData; k++) {
          TimeData3D readed = SerialMgr.parse3dStream(in);

          if (k == 0) 
              dataSet = new DataSet3D(title, t_ver, t_hn, t_port, readed);
          else 
              dataSet.addTimeData(readed);
       }

       svServer.fmanager.register(dataSet, frameID, cvalues);

    } catch (Exception e) {
      System.out.println("Receiving 3d window information error ...");
e.printStackTrace();
    }
  }

} // end of class




