
import java.awt.*;
import java.awt.event.*;
import java.util.*;

import sv.kernel.*;
import sv.util.*;
import sv.ui.*;

/** 
 * GCanvas2D.java
 * Define a canvas to draw the graph
 * @version 1.0   10/1/97
 * @author Byeongseob Ki
 */

public class GCanvas2D extends GCanvas 
            implements KeyListener, MouseListener, MouseMotionListener {

  private GFrame2D gframe;                 // parent
  svserver svServer;

  private boolean image_drawn = false;     // image drawn flag
  private static Projector projector;      // the projector

  private boolean data_available = false;  // data availability flag

  public boolean contour = false;          // contour flag

  private int prevwid, prevhgt;            // canvas size  
  private float color;                     // color of surface
  private SurfaceVertex cop;               // center of projection

  // cut and paste
  private Vector faceVec;
  private Face[] faceArr;
  private int[] findex;
  
  // setting variables
  int plot_mode;
  boolean isBoxed, isMesh, isDisplayXY, isDisplayZ, isDisplayGrids;
  private float xmin, xmax, ymin, ymax, zmin, zmax, cmin, cmax;

  // constants
  private static final int TOP         = 0;
  private static final int CENTER      = 1;


  // default font
  private Font defaultFont; 

  /**
   * The constructor of GCanvas2D
   * @param frame The parent frame. The frame must be an instance of GFrame2D
   */
  GCanvas2D(GFrame2D frame) {
    super();
    gframe = frame;
    this.svServer = gframe.svServer;

    prevwid = prevhgt = -1;
    projector = new Projector();

    projector.setDistance(gframe.config.distance_2d);
    projector.set2DScaling(gframe.config.scale_2d);       
    projector.setRotationAngle(gframe.config.theta_2d);
    projector.setElevationAngle(gframe.config.phi_2d);
    projector.set2DTranslation(gframe.config.xtrans_2d, 
                               gframe.config.ytrans_2d);

    SurfaceVertex.setProjector(projector);

    faceVec = new Vector();
    makeContourCache();

    defaultFont = new Font("TimesRoman", Font.PLAIN, 12);

    addKeyListener(this);
    addMouseMotionListener(this);
    addMouseListener(this);
  }

  
  /**
   * reset position
   */
  public void reset() {
    projector.setDistance(70);
    projector.set2DScaling(10);      
    projector.setRotationAngle(125);
    projector.setElevationAngle(10);
    projector.set2DTranslation(0,0);
    image_drawn = false;
    setup();
    repaint();
  }

  /**
   * set position
   * @param d distance
   * @param xtrans x translation
   * @param ytrans y translation
   * @param s 2d scale factor
   * @param theta rotation angle
   * @param phi elevation angle
   */
  public void setPosition(float d, int xtrans, int ytrans, 
                                float s, float theta, float phi) {
    projector.setDistance(d);
    projector.set2DTranslation(xtrans, ytrans);
    projector.set2DScaling(s);      
    projector.setRotationAngle(theta);
    projector.setElevationAngle(phi);

    gframe.config.distance_2d = d;
    gframe.config.scale_2d = s;
    gframe.config.theta_2d = theta;
    gframe.config.phi_2d = phi;
    gframe.config.xtrans_2d = xtrans;
    gframe.config.ytrans_2d = ytrans;

    image_drawn = false;
    repaint();
  }


  /**
   * Destroys the internal image. It will force GCanvas2D
   * to regenerate all images when the paint method is called. 
   */
  public void destroyImage() {
    image_drawn = false;
    faceVec.removeAllElements();
  }

  /**
   * return offscreen image
   */
  public Image getImage() {
     return offScrImage;
  }

  /** 
   * contour labels
   */
  public void drawContourLabel(boolean b) {
     isContourLabel = b;
     image_drawn = false;
     repaint();
  }

  /** 
   * set user mode contour
   * @param b collaboration
   */
  public void setContourValues(float[] values, boolean b) {
     isUserContour = true;
     contour_values = values;
     contour_lines = contour_values.length;
     clearContour();

     float min = 1.0e10f, max = -1.0e10f;
     for (int i=0; i<contour_lines; i++) {
        if (min > values[i]) min = values[i];
        if (max < values[i]) max = values[i];
     }
     gframe.setContourLines(contour_lines);
     gframe.setContourMin((float)min);
     gframe.setContourMax((float)max);
     image_drawn = false;
     repaint();
  }

  /**
   * Sets contour plot flag. GCanvas2D will generate contour
   * plot of surface if this flag is true.
   * @param contour new contour plot flag
   */
  public void setContour(boolean contour) {
    this.contour = contour;
  }


  /**
   * Sets the x and y ranges of calculated surface vertices. 
   * The ranges will not affect surface appearance. They affect axes 
   * scale appearance. 
   * @param xmin the minimum x
   * @param xmax the maximum x
   * @param ymin the minimum y
   * @param ymax the maximum y
   */
  public void setRanges(float xmin, float xmax, float ymin, float ymax) {
    this.xmin = xmin;
    this.xmax = xmax;
    this.ymin = ymin;
    this.ymax = ymax;
  }
  
    
  /**
   * Sets the data availability flag. If this flag is false,
   * GCanvas2D will not generate any surface image, even
   * if the data is available. But it is the programmer's responsiblity 
   * to set this flag to false when data is not available. 
   * @param avail the availability flag
   */
  public void setDataAvailability(boolean avail) {
    data_available = avail;
    is_data_available = avail;  // see Handlers for mouse input events section
  }


  private boolean is_data_available;   // holds the original data availability flag

  private int     click_x, click_y;    // previous mouse cursor position
    

  public void mouseClicked(MouseEvent e) {
    // NOTHING TO DO
  }

  // if mouse pressed
  public void mousePressed(MouseEvent e) {
     click_x = e.getX();
     click_y = e.getY();

     if (svServer.mergeHandler.mergeMode2D) {
        gframe.mergeAction(e.isControlDown());
     }
  }

  // if mouse released
  public void mouseReleased(MouseEvent e) {
     int x = e.getX();
     int y = e.getY();

     if (contour) {
        if (isDisplayZ) {
	   if (contour_check != null) {
	     for (int i=0; i<contour_check.length; i++) {
                if ((contour_rect[i] != null) && 
                       (contour_rect[i].contains(x, y))) {

                   String[] args = { Integer.toString(i) };
                   svServer.fmanager.action2d(gframe, 
                                   CommandSet2D.CONTOURMODIFY, args, true);
                   return;
                }
             }
           }
	}
        return;
     }

     if (isDragged) {
        String[] args = { String.valueOf(projector.getDistance()),
                          String.valueOf(projector.get2D_xTranslation()),
                          String.valueOf(projector.get2D_yTranslation()),
                          String.valueOf(projector.get2DScaling()),
                          String.valueOf(projector.getRotationAngle()),
                          String.valueOf(projector.getElevationAngle()) };
        svServer.fmanager.action2d(gframe, 
                          CommandSet2D.POSITIONCHANGE, args, true);

        if (gframe.isExpectDelay()) {
           data_available = is_data_available;

           destroyImage(); 
           repaint(); 
        }
        isDragged = false;
     }
  }


  public void mouseEntered(MouseEvent e) {
  }

  public void mouseExited(MouseEvent e) {
  }


  /**
   * mouse drag event handler. Tracks dragging operations. 
   * Checks the delay regeneration flag and does proper actions.
   */
  public void mouseDragged(MouseEvent e) {
    float new_value = 0.0f;

    int x = e.getX();
    int y = e.getY();

    if (contour) return; 
    

    if (e.isControlDown()) {
      projector.set2D_xTranslation(projector.get2D_xTranslation() 
                                                         + (x - click_x));
      projector.set2D_yTranslation(projector.get2D_yTranslation()
                                                         + (y - click_y));

    } else if (e.isShiftDown()) {

      new_value = projector.get2DScaling() + (y - click_y) * 0.5f;
      if (new_value > 60.0f) new_value = 60.0f;
      if (new_value < 2.0f) new_value = 2.0f;
      projector.set2DScaling(new_value);

    }  else {
      new_value = projector.getRotationAngle() + (x - click_x);
      while (new_value > 360) new_value -= 360;
      while (new_value < 0) new_value += 360;
      projector.setRotationAngle(new_value);
      new_value = projector.getElevationAngle() + (y - click_y);

      if (new_value > 90) 
         new_value = 90;
      else if (new_value < 0) 
         new_value = 0;

      projector.setElevationAngle(new_value);
    } 

    image_drawn = false;
    if (!gframe.isExpectDelay()) {
      isDragged = true;
      repaint(); 
    } else {
      if (!isDragged) {
        is_data_available = data_available;
        isDragged = true;
      }
      data_available = false;
      repaint();
    }        

    click_x = x;
    click_y = y;
    return;
  }

   public void mouseMoved(MouseEvent e) {
      // nothing to do
   }


  /**
   * Paints surface. Creates surface plot, contour plot, 
   * based on current vertices array and  contour plot 
   * flag. If no data is available, creates image of base plane and axes. 
   */
  public void paint(Graphics g) {
   paintDone = false;

    int hgt = getBounds().height;
    int wid = getBounds().width;

    // virtual desktop
    Rectangle r = gframe.getBounds();
    if (!r.equals(rect)) {
       if (svServer.controller != null)
          svServer.controller.desktopRepaint();
    }

    if ((wid <= 0) || (hgt <= 0)) {
       paintDone = true;
       return;
    }

    // backing offScrImage creation

    if ((wid != prevwid) || (hgt != prevhgt) || (offScrGc == null)) {

      projector.setProjectionArea(new Rectangle(0, 0, wid, hgt));  
      image_drawn = false;

      if (offScrImage != null) 
         offScrImage.flush();

      offScrImage = createImage(wid, hgt);
      if (offScrGc != null) 
         offScrGc.dispose();
      offScrGc = offScrImage.getGraphics();     

      float new_size;
      if (wid < hgt) {
         if (prevwid <= 0) new_size = 0;
         else              new_size = (float)(wid - prevwid) * 0.05f;
      } else {
         if (prevhgt <= 0) new_size = 0;
         else              new_size = (float)(hgt - prevhgt) * 0.05f;
      }
      
      float new_value = projector.get2DScaling() + new_size;
      if (new_value > 60.0f) new_value = 60.0f;
      if (new_value < 2.0f) new_value = 2.0f;
      projector.set2DScaling(new_value);


      prevwid = wid;
      prevhgt = hgt;
    }

    importVariables();
        
    printing = g instanceof PrintGraphics;
      
    if (printing) {

      // modifies variables

      Graphics savedgc = offScrGc;
      offScrGc = g;
      
      Dimension pagedimension = 
           ((PrintGraphics)g).getPrintJob().getPageDimension();
      
      printwidth = pagedimension.width;
      printheight = prevhgt * printwidth / prevwid;

      if (printheight > pagedimension.height) {
        printheight = pagedimension.height;
        printwidth = prevwid * printheight / prevhgt;
      }

      float savedscalingfactor = projector.get2DScaling();
      projector.setProjectionArea(new Rectangle(0,0,printwidth,printheight));
      projector.set2DScaling(savedscalingfactor * printwidth / prevwid);

      if (!data_available) 
        drawBoxGridsTicksLabels(offScrGc,true);
      else 
        setup();

      offScrGc.drawRect(0,0,printwidth-1,printheight-1);
      
      // restores variables
      
      projector.set2DScaling(savedscalingfactor);
      projector.setProjectionArea(new Rectangle(0,0,wid, hgt));  
      offScrGc = savedgc;
      paintDone = true;
      return;
    }
    

    if (svServer.mergeHandler.mergeMode2D && gframe.mergeSelected) {
       Font savedFont = offScrGc.getFont();
       offScrGc.setColor(Color.red);
       offScrGc.setFont(new Font("Helvetica", Font.BOLD, 24));

       offScrGc.drawString("a" + gframe.mergeIndex, wid/2, hgt/2);
       offScrGc.setFont(savedFont);
    }

    if (image_drawn && (offScrImage != null)) {
      g.drawImage(offScrImage,0,0,this);

    } else if (data_available) {
       setup();
    } else {
      g.setColor(gframe.config.bgColor);
      g.fillRect(0,0,wid,hgt); 
      displayTime(g);
      if (!contour) 
         drawBoxGridsTicksLabels(g,true);
    }

    paintDone = true;
  }

  /**
   * Updates image. Just call the paint method to avoid flickers.
   */
  public void update(Graphics g) {
    paint(g);              
  }  


/*----------------------------------------------------------------------*
 *                Private methods begin here                            *
 *----------------------------------------------------------------------*/

  private int factor_x, factor_y;   // conversion factors 
  private int t_x, t_y, t_z;        // determines ticks density


  // setup to draw 
  private void setup() {
    boolean repeat;
    int fontsize = (int)(Math.round(projector.get2DScaling() * 0.8));
    offScrGc.setFont(new Font("Helvetica", Font.PLAIN, fontsize));
    
    SurfaceVertex.invalidate(); 

    // contour plot      
    if (contour) plotContour(); 
    else {
    
    // surface plot
      if (plot_mode == gframe.WIREMODE || plot_mode == gframe.WIRECOLORMODE)
        plotWireframe();
      else if (plot_mode == gframe.POINTMODE)
        plotPointPlot();
      else
        plotSurface();
    }
    
    cleanUpMemory();
    if (printing) return;
                  
    // force repaint NOW
    Graphics g = getGraphics();
    paint(g);
    g.dispose();
  }



  /**
   * Draws the bounding box of surface.
   */

  private final void drawBoundingBox() {
    Point startingpoint, projection;

    startingpoint = projector.project(factor_x*10,factor_y*10,10);
    offScrGc.setColor(Color.black);
    projection =  projector.project(-factor_x*10,factor_y*10,10);
    offScrGc.drawLine(startingpoint.x, startingpoint.y, projection.x, projection.y);
    projection =  projector.project(factor_x*10,-factor_y*10,10);
    offScrGc.drawLine(startingpoint.x, startingpoint.y, projection.x, projection.y);
    projection =  projector.project(factor_x*10,factor_y*10,-10);
    offScrGc.drawLine(startingpoint.x, startingpoint.y, projection.x, projection.y);
  }

  // Draws the base plane. The base plane is the x-y plane.
  // parameters: x, y used to retrieve x, y coordinates 
  //             of drawn plane from this method. 
  private final void drawBase(Graphics g, int[] x, int[] y) {
    
    if (!isBoxed && !isDisplayXY && !isDisplayGrids && !isDisplayZ &&
                                      !isDragged)
        return;

    Point projection = projector.project(-10,-10,-10);                  
    x[0] = projection.x; 
    y[0] = projection.y;

    projection = projector.project(-10, 10,-10);                  
    x[1] = projection.x; 
    y[1] = projection.y;
    projection = projector.project( 10, 10,-10);                  
    x[2] = projection.x; 
    y[2] = projection.y;
    projection = projector.project( 10,-10,-10);                  
    x[3] = projection.x; 
    y[3] = projection.y;
    x[4] = x[0]; 
    y[4] = y[0];

    g.setColor(Color.black);
    g.drawPolygon(x,y,5);
  }

  /**
   * Draws non-surface parts, i.e: bounding box, axis grids, axis ticks,
   * axis labels, base plane. 
   * @param g         the graphics context to draw
   * @param draw_axes if true, only draws base plane and z axis
   */
  private final void drawBoxGridsTicksLabels(Graphics g, boolean draw_axes) {
    Point    projection, tickpos;   
    boolean  x_left = false, y_left = false;
    int      x[],y[],i;

    x = new int[5]; 
    y = new int[5];

    if (projector == null) return;

    if (draw_axes) {
      drawBase(g,x,y);
      projection = projector.project(0,0,-10);                  
      x[0] = projection.x; 
      y[0] = projection.y;
      projection = projector.project(10.5f,0,-10);                  
      g.drawLine(x[0],y[0],projection.x,projection.y);
      if (projection.x < x[0]) 
        outString(g,(int)(1.05*(projection.x-x[0]))+x[0],
                    (int)(1.05*(projection.y-y[0]))+y[0],gframe.config.xLabel,Label.RIGHT,TOP);
      else
        outString(g,(int)(1.05*(projection.x-x[0]))+x[0],
                    (int)(1.05*(projection.y-y[0]))+y[0],gframe.config.xLabel,Label.LEFT,TOP);
      projection = projector.project(0,11.5f,-10);                  
      g.drawLine(x[0],y[0],projection.x,projection.y);
      if (projection.x < x[0]) 
        outString(g,(int)(1.05*(projection.x-x[0]))+x[0],
                    (int)(1.05*(projection.y-y[0]))+y[0],gframe.config.yLabel,Label.RIGHT,TOP);
      else
        outString(g,(int)(1.05*(projection.x-x[0]))+x[0],
                     (int)(1.05*(projection.y-y[0]))+y[0],gframe.config.yLabel,Label.LEFT,TOP);
      projection = projector.project(0,0,10.5f);                  
      g.drawLine(x[0],y[0],projection.x,projection.y);
      outString(g,(int)(1.05*(projection.x-x[0]))+x[0],                    
                  (int)(1.05*(projection.y-y[0]))+y[0],gframe.config.zLabel,Label.CENTER,CENTER);
    } 
    else {
      factor_x = factor_y = 1;
      projection = projector.project(0,0,-10);                  
      x[0] = projection.x; 
      projection = projector.project(10.5f,0,-10);                  
      y_left = projection.x > x[0];
      i = projection.y;
      projection = projector.project(-10.5f,0,-10);
      if (projection.y > i) {
        factor_x = -1;
        y_left = projection.x > x[0];
      }
      projection = projector.project(0,10.5f,-10);                  
      x_left = projection.x > x[0];
      i = projection.y;
      projection = projector.project(0,-10.5f,-10);
      if (projection.y > i) {
        factor_y = -1;
        x_left = projection.x > x[0]; 
      }
      setAxesScale(); drawBase(g,x,y);

      if (isBoxed) {
        projection = projector.project(-factor_x*10,-factor_y*10,-10);
        x[0] = projection.x; 
        y[0] = projection.y;
        projection = projector.project(-factor_x*10,-factor_y*10,10);  
        x[1] = projection.x; 
        y[1] = projection.y;
        projection = projector.project(factor_x*10,-factor_y*10,10);
        x[2] = projection.x; 
        y[2] = projection.y;
        projection = projector.project(factor_x*10,-factor_y*10,-10);
        x[3] = projection.x; 
        y[3] = projection.y;
        x[4] = x[0]; 
        y[4] = y[0];
  
        g.setColor(Color.black);
        g.drawPolygon(x,y,5);

        projection = projector.project(-factor_x*10,factor_y*10,10);
        x[2] = projection.x; 
        y[2] = projection.y;
        projection = projector.project(-factor_x*10,factor_y*10,-10);
        x[3] = projection.x; 
        y[3] = projection.y;
        x[4] = x[0]; 
        y[4] = y[0];
  
        g.setColor(Color.black);
        g.drawPolygon(x,y,5);
      } else if (isDisplayZ) {
        projection = projector.project(factor_x*10,-factor_y*10,-10);
        x[0] = projection.x; 
        y[0] = projection.y;  
        projection = projector.project(factor_x*10,-factor_y*10,10);
        g.drawLine(x[0],y[0],projection.x,projection.y);

        projection = projector.project(-factor_x*10,factor_y*10,-10);
        x[0] = projection.x; 
        y[0] = projection.y;  
        projection = projector.project(-factor_x*10,factor_y*10,10);
        g.drawLine(x[0],y[0],projection.x,projection.y);
      }

      for (i=-9; i <= 9; i++) {
        if (isDisplayXY || isDisplayGrids) {
          if (!isDisplayGrids || (i % (t_y / 2) == 0) || isDisplayXY) {
            if (isDisplayGrids && (i % t_y == 0)) 
              projection = projector.project(-factor_x * 10,i,-10);
            else {
              if (i % t_y != 0) 
                projection = projector.project(factor_x * 9.8f,i,-10);
              else
                projection = projector.project(factor_x * 9.5f,i,-10);
            }
            tickpos = projector.project(factor_x * 10,i,-10);
            g.drawLine(projection.x,projection.y,tickpos.x,tickpos.y);
            if ((i % t_y == 0) && isDisplayXY) {
              tickpos = projector.project(factor_x * 10.5f,i,-10);
              if (y_left)
                outFloat(g,tickpos.x,tickpos.y,
                         (float)((float)(i+10)/20*(ymax-ymin)+ymin),
                         Label.LEFT,TOP);
              else
                outFloat(g,tickpos.x,tickpos.y,
                         (float)((float)(i+10)/20*(ymax-ymin)+ymin),
                         Label.RIGHT,TOP);
            }
          }  
          if (!isDisplayGrids || (i % (t_x / 2) == 0) || isDisplayXY) {
            if (isDisplayGrids && (i % t_x == 0))
              projection = projector.project(i,-factor_y * 10,-10);
            else {
              if (i % t_x != 0) 
                projection = projector.project(i,factor_y * 9.8f,-10);
              else
                projection = projector.project(i,factor_y * 9.5f,-10);
            }
            tickpos = projector.project(i,factor_y * 10,-10);
            g.drawLine(projection.x,projection.y,tickpos.x,tickpos.y);
            if ((i % t_x == 0) && isDisplayXY) {
              tickpos = projector.project(i,factor_y * 10.5f,-10);
              if (x_left)
                outFloat(g,tickpos.x,tickpos.y,
                         (float)((float)(i+10)/20*(xmax-xmin)+xmin),
                         Label.LEFT,TOP);
              else
                outFloat(g,tickpos.x,tickpos.y,
                         (float)((float)(i+10)/20*(xmax-xmin)+xmin),
                         Label.RIGHT,TOP);
            }
          }
        }  
        
        if (isDisplayXY) {
          tickpos = projector.project(0,factor_y * 14,-10);
          outString(g,tickpos.x,tickpos.y,gframe.config.xLabel,Label.CENTER,TOP);
          tickpos = projector.project(factor_x * 14,0,-10);
          outString(g,tickpos.x,tickpos.y,gframe.config.yLabel,Label.CENTER,TOP);
        }

        // z grids and ticks
        if (isDisplayZ || (isDisplayGrids && isBoxed)) {
          if (!isDisplayGrids || (i % (t_z / 2) == 0) || isDisplayZ) {
            if (isBoxed && isDisplayGrids && (i % t_z == 0)) {
              projection = projector.project(-factor_x*10,-factor_y*10,i);
              tickpos = projector.project(-factor_x*10,factor_y*10,i);
            }
            else {
              if (i % t_z == 0) 
                projection = projector.project(-factor_x*10,factor_y*9.5f,i);
              else
                projection = projector.project(-factor_x*10,factor_y*9.8f,i);  
              tickpos = projector.project(-factor_x*10,factor_y*10,i);
            }                            
            g.drawLine(projection.x,projection.y,tickpos.x,tickpos.y);
            if (isDisplayZ) {
              tickpos = projector.project(-factor_x*10,factor_y*10.5f,i);
              if (i % t_z == 0) {
                if (x_left)
                  outFloat(g,tickpos.x,tickpos.y,
                           (float)((float)(i+10)/20*(zmax-zmin)+zmin),
                           Label.LEFT,CENTER);
                else
                  outFloat(g,tickpos.x,tickpos.y,
                           (float)((float)(i+10)/20*(zmax-zmin)+zmin),
                           Label.RIGHT,CENTER);
              }
            }
            if (isDisplayGrids && isBoxed && (i % t_z == 0)) {
              projection = projector.project(-factor_x*10,-factor_y*10,i);
              tickpos = projector.project(factor_x*10,-factor_y*10,i);
            } 
            else {
              if (i % t_z == 0)
                projection = projector.project(factor_x*9.5f,-factor_y*10,i);
              else
                projection = projector.project(factor_x*9.8f,-factor_y*10,i);
              tickpos = projector.project(factor_x*10,-factor_y*10,i);
            }             
            g.drawLine(projection.x,projection.y,tickpos.x,tickpos.y);
            if (isDisplayZ) {
              tickpos = projector.project(factor_x*10.5f,-factor_y*10,i);
              if (i % t_z == 0) {
                if (y_left)
                  outFloat(g,tickpos.x,tickpos.y,
                           (float)((float)(i+10)/20*(zmax-zmin)+zmin),
                           Label.LEFT,CENTER);
                else
                  outFloat(g,tickpos.x,tickpos.y,
                           (float)((float)(i+10)/20*(zmax-zmin)+zmin),
                           Label.RIGHT,CENTER);
              }
            }
            if (isDisplayGrids && isBoxed) {
              if (i % t_y == 0) {
                projection = projector.project(-factor_x * 10,i,-10);
                tickpos = projector.project(-factor_x * 10,i,10);
                g.drawLine(projection.x,projection.y,tickpos.x,tickpos.y);
              }
              if (i % t_x  == 0) {
                projection = projector.project(i,-factor_y * 10,-10);
                tickpos = projector.project(i,-factor_y * 10,10);
                g.drawLine(projection.x,projection.y,tickpos.x,tickpos.y);
              }
            }   
          }
        }
      }
    }
  }

  /**
   * Imports setting variables from parent frame. This method is before doing 
   * any plotting operations
   */
  private void importVariables() {
    plot_mode = gframe.getPlotMode();
    isBoxed = gframe.isBoxed();
    isMesh = gframe.isMesh();
    isDisplayXY = gframe.isDisplayXY();
    isDisplayZ = gframe.isDisplayZ();
    isDisplayGrids = gframe.isDisplayGrids();
  }

  /**
   * Sets the axes scaling factor. Computes the proper axis lengths
   * based on the ratio of variable ranges. The axis lengths will
   * also affect the size of bounding box. 
   */

  private final void setAxesScale() {
    float scale_x, scale_y, scale_z ,divisor;
    int   longest;

    scale_x = xmax - xmin; scale_y = ymax - ymin; scale_z = zmax - zmin;

    if (scale_x < scale_y) {
      if (scale_y < scale_z) {
        longest = 3; divisor = scale_z;
      }
      else {
        longest = 2; divisor = scale_y;
      }
    }
    else {
      if (scale_x < scale_z) {
        longest = 3; divisor = scale_z;
      }
      else {
        longest = 1; divisor = scale_x;
      }
    }
    scale_x /= divisor; scale_y /= divisor; scale_z /= divisor;
 
    if ((scale_x < 0.2f) || (scale_y < 0.2f) && (scale_z < 0.2f)) {
      switch(longest) {
        case 1: if (scale_y < scale_z) {
                  scale_y /= scale_z; scale_z = 1.0f;
                }
                else {
                  scale_z /= scale_y; scale_y = 1.0f;
                } break;
        case 2: if (scale_x < scale_z) {
                  scale_x /= scale_z; scale_z = 1.0f;
                }
                else {
                  scale_z /= scale_x; scale_x = 1.0f;
                } break;
        case 3: if (scale_y < scale_x) {
                  scale_y /= scale_x; scale_x = 1.0f;
                }
                else {
                  scale_x /= scale_y; scale_y = 1.0f;
                } break;
      }
    }
    if (scale_x < 0.2f) scale_x = 1.0f; projector.setXScaling(scale_x);
    if (scale_y < 0.2f) scale_y = 1.0f; projector.setYScaling(scale_y);
    if (scale_z < 0.2f) scale_z = 1.0f; projector.setZScaling(scale_z);

    if (scale_x < 0.5f) t_x = 8; else t_x = 4;
    if (scale_y < 0.5f) t_y = 8; else t_y = 4;
    if (scale_z < 0.5f) t_z = 8; else t_z = 4;
  }

  /**
   * Draws string at the specified coordinates with the specified alignment.
   * @param g       graphics context to draw
   * @param x       the x coordinate
   * @param y       the y coordinate
   * @param s       the string to draw
   * @param x_align the alignment in x direction
   * @param y_align the alignment in y direction    
   */
  private final void outString(Graphics g, int x, int y, 
                               String s, int x_align, int y_align) {
    switch (y_align) {
      case TOP    : y += g.getFontMetrics(g.getFont()).getAscent(); break;
      case CENTER : y += g.getFontMetrics(g.getFont()).getAscent()/2; break;
    }
    switch (x_align) {
      case Label.LEFT  : 
         g.drawString(s, x, y); 
         break;

      case Label.RIGHT : 
         g.drawString(s, x-g.getFontMetrics(g.getFont()).stringWidth(s), y); 
         break;
      case Label.CENTER: 
         g.drawString(s,x-g.getFontMetrics(g.getFont()).stringWidth(s)/2,y); 
         break;
    }     
  }

  /**
   * Draws float at the specified coordinates with the specified alignment.
   * @param g       graphics context to draw
   * @param x       the x coordinate
   * @param y       the y coordinate
   * @param f       the float to draw
   * @param x_align the alignment in x direction
   * @param y_align the alignment in y direction    
   */
  private final void outFloat(Graphics g, int x, int y, 
                              float f, int x_align, int y_align) {
    String s =  SVUIC.formatFloat(f, 3, 2, SVUIC.ALGEBRAIC_FORMAT);
    outString(g,x,y,s,x_align,y_align);
  }


  public void keyTyped(KeyEvent e) {
  }

  public void keyReleased(KeyEvent e) {
  }

  public void keyPressed(KeyEvent e) {
      if (e.isControlDown()) {   
        switch(e.getKeyCode()) {
           case 65:                       // ctrl-a animation
              svServer.fmanager.action2d(gframe, CommandSet2D.ANIMATE, 
                                                              null, true);
              break;

           case 83:                       // ctrl-s stop
              String[] arg = new String[1];
              arg[0] = String.valueOf(gframe.currentTime);
              svServer.fmanager.action2d(gframe, CommandSet2D.STOP, 
                                                              arg, true);
              break;

           case 78:                       // ctrl-n next
              svServer.fmanager.action2d(gframe, CommandSet2D.NEXT, 
                                                              null, true);
              break;

           case 80:                      // ctrl-p previous
              svServer.fmanager.action2d(gframe, CommandSet2D.PREVIOUS, 
                                                              null, true);
              break;

           case 81:                      // ctrl-q kill window
              svServer.fmanager.action2d(gframe, CommandSet2D.QUIT, 
                                                              null, true);
              break;

           case 71:                      // ctrl-g go specific time
              gframe.doWork(CommandSet2D.GOTO, null);
              break;

           case 82:                      // ctrl-r rename dialog
              gframe.doWork(CommandSet2D.RENAMETITLE, null);
              break;
        }
      }
  }


/*--------------------------------------------------------------------*
 *        Plotting routines and methods begin here                    *
 *--------------------------------------------------------------------*/
  private float color_factor;
  private Point projection;   
  private Color line_color; 
  
  private final int poly_x[] = new int[9];                       
  private final int poly_y[] = new int[9];                   

  /**
   * Plots a single plane
   * @param vertex vertices array of the plane
   * @param verticescount number of vertices to process 
   * @param color color vlaue
   * @param isCustom is cumstom color?
   */
  private final void plotPlane(SurfaceVertex[] vertex, int verticescount, float color, boolean isCustom) {
    int     loop, index;
    float   result;
    boolean low1, low2;
    boolean valid1, valid2;

    if (verticescount < 3) return; 

    int count = 0; 
    float z = 0.0f; 

    line_color = Color.black;
    low1 = (vertex[0].z < zmin); 
    valid1 = !low1 && (vertex[0].z <= zmax); 
    index = 1;

    for (loop = 0; loop < verticescount; loop++) {
      low2 = (vertex[index].z < zmin); valid2 = !low2 && (vertex[index].z <= zmax);
      if ((valid1 || valid2) || (low1^low2)) {
        if (!valid1) {
          if (low1) result = zmin; else result = zmax;
          float ratio = (result-vertex[index].z)/(vertex[loop].z-vertex[index].z);
          float new_x = ratio*(vertex[loop].x-vertex[index].x)+vertex[index].x; 
          float new_y = ratio*(vertex[loop].y-vertex[index].y)+vertex[index].y;
          if (low1)
            projection = projector.project(new_x,new_y,-10);
          else
            projection = projector.project(new_x,new_y,10);
          poly_x[count] = projection.x;
          poly_y[count] = projection.y;  
          count++; z += result;
        } 
        if (valid2) {
          projection = vertex[index].projection();
          poly_x[count] = projection.x;
          poly_y[count] = projection.y;  
          count++; z += vertex[index].z;
        }
        else {
          if (low2) result = zmin; 
          else result = zmax;
          float ratio = 
              (result-vertex[loop].z)/(vertex[index].z-vertex[loop].z);
          float new_x = ratio*(vertex[index].x-vertex[loop].x)+vertex[loop].x; 
          float new_y = ratio*(vertex[index].y-vertex[loop].y)+vertex[loop].y;

          if (low2)
            projection = projector.project(new_x,new_y,-10);
          else
            projection = projector.project(new_x,new_y,10);

          poly_x[count] = projection.x;
          poly_y[count] = projection.y;  
          count++; z += result;
        } 
      }
      if (++index == verticescount) index = 0; 
      valid1 = valid2; low1 = low2;
    }
    if (count > 0) {
      switch (plot_mode) {
        case gframe.HIDDENMODE:    offScrGc.setColor(Color.lightGray); 
                          break;
        case gframe.SPECTRUMMODE:   
          int val = 0;

	  if (isCustom) 
            val = (int)((color - cmin) / (cmax-cmin) * 100f);
          else 
            val = (int)((z/count - zmin) / (zmax-zmin) * 100f);

          if (gframe.colorModel == GFrame.RGBCOLORMODEL) 
             offScrGc.setColor(
               new Color((int)(gframe.rcolors[val] * 255f), 
                         (int)(gframe.gcolors[val] * 255f), 
                         (int)(gframe.bcolors[val] * 255f)));
          else 
              offScrGc.setColor(
                Color.getHSBColor((float)gframe.hcolors[val], 
                                  (float)gframe.scolors[val],
                                  (float)gframe.lcolors[val]));

          break;

        case gframe.GRAYMODE:   z = (z/count-zmin)*color_factor;
                          offScrGc.setColor(Color.getHSBColor(0,0,z));
                          if (z < 0.3f) line_color = new Color(0.6f,0.6f,0.6f);                    
                          break; 
      } 
                               
      offScrGc.fillPolygon(poly_x, poly_y, count);
      offScrGc.setColor(line_color);
      if (isMesh || (plot_mode == gframe.HIDDENMODE)) {
        poly_x[count] = poly_x[0];
        poly_y[count] = poly_y[0];
        count++;
        offScrGc.drawPolygon(poly_x, poly_y, count);
      }
    }
  }
  
  
  /**
   * Determines whether a plane is plottable, i.e: does not have
   * invalid vertex.
   * @return true if the plane is plottable, false otherwise
   * @param values vertices array of the plane
   */
  private final static boolean plottable(SurfaceVertex[] values) {
    return (!values[0].isInvalid() && !values[1].isInvalid() &&
            !values[2].isInvalid() && !values[3].isInvalid()); 
  }

  // Plots an area of group of planes 
  // it works only the number of data set size is 1
  private final void plotArea(int start_lx, int start_ly, 
                          int end_lx, int end_ly, int sx, int sy) {

    // start_lx, slart_ly  start index in x and y direction
    // end_lx, end_ly   end index in x and y direction
    // sx, sy step in x and y direction

    DataSet2D data = (DataSet2D)gframe.dataSetVector.firstElement();
    int dimX = data.timeData[gframe.currentTime].getDimX() - 1;
    int dimY = data.timeData[gframe.currentTime].getDimY() - 1;
    SurfaceVertex[] vertex = data.timeData[gframe.currentTime].vertex;
    boolean isCustom = false;
    if (data.timeData[gframe.currentTime].cArr != null) isCustom = true;
 
    start_lx *= dimY + 1;
    sx *= dimY + 1;
    end_lx *= dimY + 1;
    int lx = start_lx;
    int ly = start_ly;

    SurfaceVertex values[] = new SurfaceVertex[4]; 

    while (ly != end_ly) {
      values[1] = vertex[lx+ly];
      values[2] = vertex[lx+ly+sy];
        
      while (lx != end_lx) {
        values[0] = values[1];
        values[1] = vertex[lx+sx+ly];
        values[3] = values[2];
        values[2] = vertex[lx+sx+ly+sy];

        if (plottable(values)) {
           if (!isCustom) plotPlane(values, 4, -1.0f,  false);
           else plotPlane(values, 4, data.timeData[gframe.currentTime].cArr[lx+ly], true);
        }
        lx += sx; 
      }
      ly += sy; lx = start_lx;
    }
  }

  // make an object to calculate the z-depth
  private final void makeObject(int start_lx, int start_ly, 
       int end_lx, int end_ly, int sx, int sy, DataSet2D data, int pos) {

    // start_lx, slart_ly  start index in x and y direction
    // end_lx, end_ly   end index in x and y direction
    // sx, sy step in x and y direction

    int dimX = data.timeData[pos].getDimX() - 1;
    int dimY = data.timeData[pos].getDimY() - 1;
    SurfaceVertex[] vertex = data.timeData[pos].vertex;

    start_lx *= dimY + 1;
    sx *= dimY + 1;
    end_lx *= dimY + 1;
    int lx = start_lx;
    int ly = start_ly;

    SurfaceVertex values[] = new SurfaceVertex[4]; 

    while (ly != end_ly) {
      values[1] = vertex[lx+ly];
      values[2] = vertex[lx+ly+sy];
        
      while (lx != end_lx) {
        values[0] = values[1];
        values[1] = vertex[lx+sx+ly];
        values[3] = values[2];
        values[2] = vertex[lx+sx+ly+sy];

        faceVec.addElement(new Face(values));
        lx += sx; 
      }
      ly += sy; lx = start_lx;
    }
  }

    
  /**
   * Creates a surface plot
   */
  private final void plotSurface() {
    float    zi, zx;
    int      sx, sy;
    int      start_lx, end_lx;
    int      start_ly, end_ly;
    
    image_drawn = false; 
    try {
      zi = gframe.getZMin();
      zx = gframe.getZMax();
      cmin = gframe.getCMin();
      cmax = gframe.getCMax();
      if (zi > zx || cmin > cmax) throw new NumberFormatException();
      
      if (zi == zx) 
          zi = zi - 1.0f;
      if (cmin == cmax) 
          cmin = cmin - 1.0f;
    } catch(NumberFormatException e) {
      gframe.monitor.display("Error in ranges", true); 
      return;
    }

    zmin = zi; zmax = zx; 
    color_factor = 0.8f / (zmax-zmin);    

    offScrGc.setColor(gframe.config.bgColor);
    offScrGc.fillRect(0,0,getBounds().width,getBounds().height);
    displayTime(offScrGc);
    
    drawBoxGridsTicksLabels(offScrGc,false); 

    if (gframe.dataSetVector.size() < 1) {    
      if (isBoxed) drawBoundingBox();

      image_drawn = true;
      return; 
    }

    SurfaceVertex.setZRange(zmin,zmax);

    // direction test
    float distance = projector.getDistance() * 
                       projector.getCosElevationAngle();

    // cop : center of projection
    cop = new SurfaceVertex(distance * projector.getSinRotationAngle(), 
               distance * projector.getCosRotationAngle(), 
               projector.getDistance() * projector.getSinElevationAngle());
    cop.transform();

    boolean inc_x = cop.x > 0;
    boolean inc_y = cop.y > 0;
    

    if (gframe.dataSetVector.size() == 1) {    

       DataSet2D data = (DataSet2D)gframe.dataSetVector.firstElement();
       int dimX = data.timeData[gframe.currentTime].getDimX() - 1;
       int dimY = data.timeData[gframe.currentTime].getDimY() - 1;

       if (inc_x) {
         start_lx = 0; end_lx = dimX; 
         sx = 1;
       } else {
         start_lx = dimX; end_lx = 0; 
         sx = -1; 
       } 
    
       if (inc_y) {
         start_ly = 0; end_ly = dimY; 
         sy = 1; 
       } else {
         start_ly = dimY; end_ly = 0; 
         sy = -1; 
       }

       if ((cop.x > 10) || (cop.x < -10)) {
         if ((cop.y > 10) || (cop.y < -10)) {
           plotArea(start_lx, start_ly, end_lx, end_ly, sx, sy);
         } else {    // split in y direction
           int split_y = (int)(cop.y+10);
           if (split_y > dimY)
              plotArea(start_lx, start_ly, end_lx, end_ly, sx, sy);
           else {
              plotArea(start_lx, 0, end_lx, split_y, sx, 1);
              plotArea(start_lx, dimY, end_lx, split_y, sx, -1);
           }
         }
       } else {
         if ((cop.y > 10) || (cop.y < -10)) {   // split in x direction
           int split_x = (int)(cop.x+10);
           if (split_x > dimX)
              plotArea(start_lx, start_ly, end_lx, end_ly, sx, sy);
           else {
              plotArea(0, start_ly, split_x, end_ly, 1, sy);
              plotArea(dimX, start_ly, split_x, end_ly, -1, sy);   
           }     
         } else {    // split in both x and y directions
           int split_x = (int)(cop.x+10);   
           int split_y = (int)(cop.y+10); 
           if ((split_x > dimX) || (split_y > dimY))
              plotArea(start_lx, start_ly, end_lx, end_ly, sx, sy);
           else {
              plotArea(0, 0, split_x, split_y, 1, 1);      
              plotArea(0, dimY, split_x, split_y, 1, -1); 
              plotArea(dimX, 0, split_x, split_y, -1, 1);  
              plotArea(dimX, dimY, split_x, split_y, -1, -1);      
           }
         }
       }
       gframe.monitor.display((gframe.currentTime+1) + " / " + 
         gframe.noOfTime + "   ( " + (dimX+1) +"x"+ (dimY+1) + " )", false);

    ////////////////////////////////////////////////////////////////
    } else {    // if the number of data set is larger than 1

       faceVec.removeAllElements();

       for (int i=0; i<gframe.dataSetVector.size(); i++) {    
         
          DataSet2D data = (DataSet2D)gframe.dataSetVector.elementAt(i);

          for (int k=0; k<data.getTimeArr().length; k++) {
             if (data.getTimeValue(k) == gframe.timeArr[gframe.currentTime]) {

               int dimX = data.timeData[k].getDimX() - 1;
               int dimY = data.timeData[k].getDimY() - 1;

               makeObject(0, 0, dimX, dimY, 1, 1, data, k);
             }
	  }
       }

       faceArr = new Face[faceVec.size()];
       findex = new int[faceVec.size()];

       for (int i=0; i<faceArr.length; i++) {
          faceArr[i] = (Face)faceVec.elementAt(i);
          faceArr[i].calc_zdepth(cop);
          findex[i] = i;
       }

       qs(0, faceArr.length-1);
       for (int i=0; i<faceArr.length; i++) 
          plotPlane(faceArr[findex[i]].vertex, 4, -1.0f, false);

       gframe.monitor.display((gframe.currentTime+1) + " / " + 
                                           gframe.noOfTime, false);


    }
    ////////////////////////////////////////////////////////////////


    if (isBoxed) 
       drawBoundingBox();

    image_drawn = true; 
  }


 /**
  *  The quick sort algorithm in this method is used for sorting faces
  *  from back to front by z depth values.
  */
  void qs(int left, int right) {

    int i, j, k;
    float x;

    i = left; j = right;
    x = faceArr[findex[(left+right)/2]].zdepth;

    do {

      while (faceArr[findex[i]].zdepth > x && i < right) i++;
      while (x > faceArr[findex[j]].zdepth && j > left) j--;

      if (i <= j) {
        k = findex[i];
        findex[i] = findex[j];
        findex[j] = k;
        i++; j--;
      }

    } while (i <= j);

    if (left < j) qs(left, j);
    if (i < right) qs(i, right);

  }

  
  private int      contour_center_x  = 0;
  private int      contour_center_y  = 0;
  private int      contour_space_x   = 0;
  private int      legend_width      = 0;
  private int      legend_space      = 0;
  private int      legend_length     = 0;
  private String[] legend_label      = null;

  private float    contour_width_x   = 0.0f;
  private float    contour_width_y   = 0.0f;

  private String[] ylabels           = null;
  
  private int[]    xpoints           = new int[8];
  private int[]    ypoints           = new int[8];

  private int[]    contour_x         = new int[8];
  private int[]    contour_y         = new int[8];
  private int      contour_n         = 0;
    
  private int      contour_lines     = 10;
  private Color[]  contour_color     = null;

  private Rectangle[] contour_rect   = null;
  private float[]    contour_values = null;
  private boolean[]   contour_check  = null;
  private boolean     contour_recalc = true;

  private float contour_stepz;
  private SurfaceVertex[] ContourVtx = new SurfaceVertex[4]; 

  private boolean isContourLabel = false;  
  private boolean isUserContour = false;


  // return contour information by string
  private String getContourInfo() {
     String info = Integer.toString(contour_lines) + ";";

     for (int i=0; i<contour_lines; i++) 
        info += Float.toString(contour_values[i]) + ";";

      return info;
  }


  // modify contour
  public void modifyContour(int i) {
     if (contour) {
       if (isDisplayZ) {
          if (contour_check != null) {
	     if (contour_rect != null) {
                 if (contour_check[i] == true) contour_check[i] = false;
                 else contour_check[i] = true;

                 makeContourCache();
                 destroyImage();
                 repaint();
              }
           }
         }
      }
   }

  /**
   * Converts normalized x coordinate (-10..+10) to screen x coordinate 
   * @return the screen x coordinate
   * @param x the normalized x coordinate
   */
  private final int contourConvertX(float x) {
    return (int)Math.round(x * contour_width_x + contour_center_x);
  }
  
  /**
   * Converts normalized y coordinate (-10..+10) to screen y coordinate 
   * @return the screen y coordinate
   * @param y the normalized x coordinate
   */
  private final int contourConvertY(float y) {
    return (int)Math.round(-y * contour_width_y + contour_center_y);
  }

  /**
   * Creates bounding box for images of contour plot 
   */ 
  private final void drawBoundingRect() {
    offScrGc.setColor(Color.black);
    int x1 = contourConvertX(-10);
    int y1 = contourConvertY(+10);
    int x2 = contourConvertX(+10);
    int y2 = contourConvertY(-10);        
    offScrGc.drawRect(x1, y1, x2-x1, y2-y1);

    if (isDisplayXY || isDisplayGrids) {    
      
      if (isDisplayXY) {
        x1 = contourConvertX(-10.5f);
        y1 = contourConvertY(+10.5f);
        x2 = contourConvertX(+10.5f);
        y2 = contourConvertY(-10.5f);        
        offScrGc.drawRect(x1, y1, x2-x1, y2-y1);
      }
      
      int xc,yc,labelindex = 0;
    
      for (int i=-10; i <= 10; i++) {
        if (!isDisplayGrids || (i % (t_y / 2) == 0) || isDisplayXY) {
          yc = contourConvertY(i);          
          if ((isDisplayGrids) && (i % t_y == 0)) { 
            offScrGc.drawLine(contourConvertX(-10.0f),yc,
                              contourConvertX(+10.0f),yc);
          }
          if (isDisplayXY) {
            if (i % t_y != 0) {
              offScrGc.drawLine(contourConvertX(+10.3f),yc,x2,yc);
              offScrGc.drawLine(contourConvertX(-10.3f),yc,x1,yc);
            }
            else {              
              offScrGc.drawLine(contourConvertX(+10.0f),yc,x2,yc);
              offScrGc.drawLine(contourConvertX(-10.0f),yc,x1,yc);
            }
          }
          if ((i % t_y == 0) && isDisplayXY) {
            outString(offScrGc,contourConvertX(10.7f),yc,
                      ylabels[labelindex++],Label.LEFT,CENTER);
          }
        }  
        if (!isDisplayGrids || (i % (t_x / 2) == 0) || isDisplayXY) {
          xc = contourConvertX(i);
          if ((isDisplayGrids) && (i % t_x == 0)) {
            offScrGc.drawLine(xc,contourConvertY(-10.0f),
                              xc,contourConvertY(+10.0f));
          }
          if (isDisplayXY) {
            if (i % t_x != 0) {
              offScrGc.drawLine(xc,contourConvertY(-10.3f),xc,y2);
              offScrGc.drawLine(xc,contourConvertY(+10.3f),xc,y1);
            }
            else {
              offScrGc.drawLine(xc,contourConvertY(-10.0f),xc,y2);
              offScrGc.drawLine(xc,contourConvertY(+10.0f),xc,y1);
            }
          }
          if ((i % t_x == 0) && isDisplayXY) {
            outFloat(offScrGc,xc,contourConvertY(-10.7f),
                     (float)((float)(i+10)/20*(xmax-xmin)+xmin),
                     Label.CENTER,TOP);
          }        
        }          
        if (isDisplayXY) {
          outString(offScrGc,(x1+x2)/2,
                    contourConvertY(-11.4f),gframe.config.xLabel,Label.CENTER,TOP);
          outString(offScrGc,contourConvertX(10.7f),
                    contourConvertY(-1.0f),gframe.config.yLabel,Label.LEFT,CENTER);
        }
      }
    }

    if (isDisplayZ) {
      int lasty = y2, height = y2-y1;
      int divisions = contour_lines;

      x2 += contour_space_x;
      x1 = x2 - (legend_space + legend_width + legend_length);
      x2 -= legend_length;
      
      //      offScrGc.setColor(Color.black);
      //      outString(offScrGc,x2,y2,legend_label[0],Label.LEFT,CENTER);
           
      for (int i=0; i < divisions; i++) {
        int y = y2 - (int)((i+1)*height/divisions);
        if ((plot_mode == gframe.SPECTRUMMODE) || (plot_mode == gframe.GRAYMODE))
           offScrGc.setColor(contour_color[i]);
        else
           offScrGc.setColor(Color.white);
        contour_rect[i] = new Rectangle( x1, y, legend_width, lasty-y);
        offScrGc.fillRect(contour_rect[i].x, contour_rect[i].y, 
              contour_rect[i].width, contour_rect[i].height);
        offScrGc.setColor(Color.black);
        offScrGc.drawRect(contour_rect[i].x, contour_rect[i].y, 
              contour_rect[i].width, contour_rect[i].height);

        if (contour_check[i]) {
            if (plot_mode == gframe.SPECTRUMMODE) 
               offScrGc.setColor(Color.white);
            else
               offScrGc.setColor(Color.red);
            offScrGc.drawLine(contour_rect[i].x, contour_rect[i].y, 
              contour_rect[i].x+contour_rect[i].width, 
              contour_rect[i].y+contour_rect[i].height);
            offScrGc.drawLine(contour_rect[i].x, 
              contour_rect[i].y+contour_rect[i].height, 
              contour_rect[i].x+contour_rect[i].width, 
              contour_rect[i].y);
            offScrGc.setColor(Color.black);
        }
        outString(offScrGc, x2, y+5, legend_label[i], Label.LEFT,CENTER);
        lasty = y;
      }
    }    
  }

  /**
   * Common method for contour plot 
   * cleans up memory used to hold temporary variables. 
   */
  private final void cleanUpMemory() {
    legend_label = null;
    ylabels = null;
  }


  /**
   * clear contour
   */
  public void clearContour() {
     contour_rect = null;
     contour_check = null;
     contour_recalc = true;
     makeContourCache();
  }
  
  /**
   * Common method for contour plot 
   * computes the best plot area size and position.
   */
  private final void computePlotArea() {
    setAxesScale();
    contour_lines = gframe.getContourLines();

    float ratio = projector.getYScaling() / projector.getXScaling();
        
    float width = 0.0f;
    float height = 0.0f;
    
    if (printing) {
      width = printwidth;
      height = printheight;
    }
    else {
      width = getBounds().width;
      height = getBounds().height;
    }

    int fontsize = 0;
    
    if (width < height) fontsize = (int)(width / 48);
                   else fontsize = (int)(height / 48);
    offScrGc.setFont(new Font("Helvetica",Font.PLAIN,fontsize));

    FontMetrics fm = offScrGc.getFontMetrics();
    
    // Leaves boundary space
     
    width *= 0.9f;
    height *= 0.9f;
    
    int spacex = 0;
    int spacey = 0;
    
    if (isDisplayXY) {

      // Creates y labels bank
      
      int labelscount = 0, index = 0, maxwidth = 0;
      for (int i=-10; i < 10; i++) if (i % t_y == 0) labelscount++;
      ylabels = new String[labelscount];

      for (int i=-10; i < 10; i++) {
        if (i % t_y == 0) {
          ylabels[index] = 
          new String(Float.toString(
                    (float)((float)(i+10)/20*(ymax-ymin)+ymin)));
          int strwidth = fm.stringWidth(ylabels[index++]);
          if (strwidth > maxwidth) maxwidth = strwidth;
        }
      }

      spacex += maxwidth;
      spacey += fm.getMaxAscent();
    }

    if (isDisplayZ) {

        legend_width = (int)(width * 0.05);
        spacex += legend_width * 2;
        legend_space = fontsize;
        
        // Create Z Labels
        

        legend_length = 0;
        legend_label = new String[contour_lines];

        for (int i=0; i < contour_lines; i++) {
          float label;

          if (isUserContour) {
             label = contour_values[i];
          } else
             label = (float)((float)(i)/contour_lines*(zmax-zmin)+zmin);

          legend_label[i] = new String(Float.toString(label));
          int labelwidth = fm.stringWidth(legend_label[i]);
          if (labelwidth > legend_length) legend_length = labelwidth;
        }
        
        spacex += legend_length + legend_space;
	//      }
    } else {
      legend_length = 0;
      legend_label = new String[contour_lines];
      for (int i=0; i < contour_lines; i++) {
        float label = (float)((float)(i)/(contour_lines-1)*(zmax-zmin)+zmin);
        legend_label[i] = new String(Float.toString(label));
        int labelwidth = fm.stringWidth(legend_label[i]);
        if (labelwidth > legend_length) legend_length = labelwidth;
      }
    }
        
    width -= spacex;
    height -= spacey;
    
    contour_width_x = width;
    contour_width_y = width * ratio;

    if (contour_width_y > height) {
      contour_width_y = height;
      contour_width_x = height / ratio;
    }
        
    float scaling_factor = 10.0f;
    if (isDisplayXY) scaling_factor = 10.7f; // + 1.4 / 2
     
    contour_width_x = contour_width_x / scaling_factor / 2;
    contour_width_y = contour_width_y / scaling_factor / 2;
        
    contour_center_x = 0;
    contour_center_y = 0;

    int x1 = contourConvertX(-scaling_factor);       
    int y1 = contourConvertY(+scaling_factor);       
    int x2 = contourConvertX(+scaling_factor) + spacex;       
    int y2 = contourConvertY(-scaling_factor) + spacey;
    
    // computes center
    
    contour_center_x = (getBounds().width-(x1+x2))/2;       
    contour_center_y = (getBounds().height-(y1+y2))/2;       

    contour_space_x = spacex;

    contour_rect = new Rectangle[contour_lines];
    
    if (contour_recalc) {
       contour_check = new boolean[contour_lines];
       for (int i=0; i<contour_lines; i++)
         contour_check[i] = false;
       contour_recalc = false;
    }

    // Creates color bank
    contour_color = new Color[contour_lines];
    for (int i=0; i < contour_lines; i++) {
      float level = (float)i / contour_lines * 0.8f;
      int val = (int)((float)i / (float)contour_lines * 100f);

      switch (plot_mode) {
        case gframe.SPECTRUMMODE:
            if (gframe.colorModel == GFrame.RGBCOLORMODEL) 
               contour_color[i] = new Color((int)(gframe.rcolors[val] * 255f),
                                         (int)(gframe.gcolors[val] * 255f),
                                         (int)(gframe.bcolors[val] * 255f));
            else
               contour_color[i] = 
                   Color.getHSBColor((float)gframe.hcolors[i], 
                                     (float)gframe.scolors[i],
                                     (float)gframe.lcolors[i]);

            break;
        default:          
            contour_color[i] = Color.getHSBColor(0,0,level);
            break; 
      }
    }
  }
   
  // calculate contour lines
  private final ContourData calcContour(TimeData2D timeData) {
    
     float[] grid = timeData.getzArr();
     int dimX = timeData.getDimX() - 1;
     int dimY = timeData.getDimY() - 1;

     IsoCurve isocurve = new IsoCurve(grid, dimX+1, dimY+1);
     ContourData cdata = new ContourData(contour_lines);
     float isodata[];

     float pinc = 100/contour_lines, pinc2;
     int progress;

     for(int i=0; i<contour_lines; i++) {
        if (!contour_check[i]) {
          float level;
          if (isUserContour) {
             level = contour_values[i];
          } else
             level = (float)((float)(i)/contour_lines*(zmax-zmin)+zmin);

	  //          System.out.println("Calculating Contours: level="+level);
          isocurve.setValue(level);
       
          Vector datavec = new Vector();
          while( (isodata = isocurve.getCurve()) != null ) {
             datavec.addElement(new ContourLineData(isodata));
             pinc2 = pinc / isodata.length;
             for (int j=0; j<isodata.length-3; j=j+2) {
                progress = (int)((float)i*pinc + (float)j*pinc2);
                gframe.monitor.display(Integer.toString(progress) + 
                                                  "% calculated", false);
              }
	  }
          cdata.addContourLine(level, datavec);
	} else {
          cdata.addContourLine(0.0f, null);
        }
     }
     isocurve = null;
     return cdata;
  }

  // for contour cache
  private ContourData[] contourCacheArr;

  /**
   * make contour cache
   */
  public void makeContourCache() {
    //System.out.println("Reconstruct contour cache ..");
    if (gframe.dataSetVector.size() < 1) return;

    DataSet2D data = (DataSet2D)gframe.dataSetVector.firstElement();
    contourCacheArr = new ContourData[data.getTimeArr().length];
    System.gc();
  }

  // Creates contour plot
  private final void plotContour() {
    float zi, zx;
    image_drawn = false; 
    
    try {
      zi = gframe.getContourMin();
      zx = gframe.getContourMax();
      if (zi > zx) throw new NumberFormatException();
      if (zi == zx) 
          zi = zi - 0.0001f;
    } catch(NumberFormatException e) {
      gframe.monitor.display("Error in ranges", true); 
      return;
    }

    zmin = zi; zmax = zx; 
    computePlotArea();
    
    offScrGc.setColor(Color.white);
    offScrGc.fillRect(0,0,getBounds().width,getBounds().height);
    displayTime(offScrGc);

    // Contour lines    
    offScrGc.setColor(Color.black);    
    
    // Bounding rectangle    
    drawBoundingRect();

    
    ////////////////////////////////////////////////////////
    // calculate contour lines
    ////////////////////////////////////////////////////////
    DataSet2D data = (DataSet2D)gframe.dataSetVector.firstElement();

    for (int kl=0; kl<data.getTimeArr().length; kl++) {
       if (data.getTimeValue(kl) == gframe.timeArr[gframe.currentTime]) {
       
          float x1, x2, y1, y2;

          ContourData cdata;
          if (contourCacheArr[kl] == null) {
             cdata = calcContour(data.timeData[kl]);
             contourCacheArr[kl] = cdata;
          } else
             cdata = contourCacheArr[kl];

          if (cdata.getCount() != contour_lines) {
            gframe.monitor.display("Contour calculation error ...", true);
          }

          int dimX = data.timeData[kl].getDimX() - 1;
          int dimY = data.timeData[kl].getDimY() - 1;

          for(int i=0; i<contour_lines; i++) {
            Vector dataVec = cdata.getData(i);

            if (dataVec == null) continue;

            if ((plot_mode == gframe.SPECTRUMMODE) || (plot_mode == gframe.GRAYMODE))
               offScrGc.setColor(contour_color[i]);
            else
               offScrGc.setColor(Color.black);

            for (Enumeration e=dataVec.elements(); e.hasMoreElements();) {
                ContourLineData linedata = (ContourLineData)e.nextElement();

                float[] isodata = linedata.getContourLineData();

                   for (int j=0; j<isodata.length-3; j=j+2) {
                     x1 = (float)isodata[j] * 20f / (float)dimX - 10f;
                     y1 = (float)isodata[j+1] * 20f / (float)dimY - 10f;
                     x2 = (float)isodata[j+2] * 20f / (float)dimX - 10f;
                     y2 = (float)isodata[j+3] * 20f / (float)dimY- 10f;
                     if (x1 > -10f && x1 < 10f && x2 > -10f && x2 < 10f &&
                          y1 > -10f && y1 < 10f && y2 > -10f && y2 < 10f)
                          offScrGc.drawLine(
                             contourConvertX(x1), contourConvertY(y1),
                             contourConvertX(x2), contourConvertY(y2));
                  }

                  if (isContourLabel) {
                     int idx = (int)(Math.random()*100d);
                     if ((isodata.length > idx+1) && (idx > 1)) {
                        FontMetrics fm = offScrGc.getFontMetrics();
                        Rectangle r = new Rectangle();

                        r.width  = fm.stringWidth(legend_label[i]);
                        r.height = fm.getAscent();
                        if ((idx % 2) == 0) {
                           r.x = 
contourConvertX((float)isodata[idx] * 20f / (float)dimX - 10f) - r.width/2;
                           r.y = 
contourConvertY((float)isodata[idx+1] * 20f / (float)dimX - 10f) - r.height/2;
                        } else {
                           r.x = 
contourConvertX((float)isodata[idx-1] * 20f / (float)dimX - 10f) - r.width/2;
                           r.y = 
contourConvertY((float)isodata[idx] * 20f / (float)dimX - 10f) - r.height/2;
                        }
                        offScrGc.setColor(Color.white);
                        offScrGc.fillRect(r.x, r.y, r.width, r.height);
                        if ((plot_mode == gframe.SPECTRUMMODE) || (plot_mode == gframe.GRAYMODE))
                           offScrGc.setColor(contour_color[i]);
                        else
                           offScrGc.setColor(Color.black);
                        offScrGc.drawString(legend_label[i], r.x, r.y+r.height);
		     }
		  }
	    }
	  }

          // draw vector set
          if (gframe.isVectorPlot()) 
            drawVectorPlot();
    
         gframe.monitor.display((gframe.currentTime+1) + " / " 
    + gframe.noOfTime + "   ( " + (dimX+1) +"x"+ (dimY+1) + " )", false);
       }
    }
    image_drawn = true; 
  }


  /**
   * Draw vector plot 
   */
  private final void drawVectorPlot() {
    /********************************************************
     * SHOULD BE IMPLEMENT LATER
    DataSet2D dataSet = (DataSet2D)gframe.dataSetVector.firstElement();

    for (int k=0; k<dataSet.getTimeArr().length; k++) {
      if ( (dataSet.getTimeValue(k) == gframe.timeArr[gframe.currentTime])
         && (dataSet.timeData[k].getVectorSet() != null) ) {
       TimeDataVec vs = 
             dataSet.timeData[k].getVectorSet();

       int size = vs.xArr.length;

       int xx, yy;
       int ix, iy;
       for (int i=0; i<size; i++) {
          float x = (float)(vs.xArr[i] + vs.rArr[i] * Math.cos(vs.tArr[i]));
          float y = (float)(vs.yArr[i] + vs.rArr[i] * Math.sin(vs.tArr[i]));

          if ((x < -10.0f) || (x > 10.0f) || (y < -10.0f) || (y > 10.0f))
              continue;

          xx = contourConvertX((float)vs.xArr[i]);
          yy = contourConvertY((float)vs.yArr[i]);
          ix = contourConvertX(x);
          iy = contourConvertY(y);

          offScrGc.drawLine(xx, yy, ix, iy);

          float dx = ix - xx;
          float dy = iy - yy;

          int ix0 = ix - (int)(0.25*(dx-dy));
          int iy0 = iy - (int)(0.25*(dx+dy)); 
          offScrGc.drawLine(ix0, iy0, ix, iy);

          ix0 = ix - (int)(0.25*(dx+dy));
          iy0 = iy - (int)(0.25*(-dx+dy)); 
          offScrGc.drawLine(ix0, iy0, ix, iy);
       }
     }
    }
    */
  }


  // Creates point plot 
  private final void plotPointPlot() {
    float    zi,zx;    
    float    z;
    boolean  invalid;
      
    image_drawn = false; projection = new Point(0,0); 

    try {
      zi = gframe.getZMin();
      zx = gframe.getZMax();
      cmin = gframe.getCMin();
      cmax = gframe.getCMax();
      if (zi > zx || cmin > cmax) throw new NumberFormatException();
      
      if (zi == zx) 
          zi = zi - 1.0f;
      if (cmin == cmax) 
          cmin = cmin - 1.0f;
    } catch(NumberFormatException e) {
      gframe.monitor.display("Error in ranges", true); 
      return;
    }

    zmin = zi; 
    zmax = zx; 

    offScrGc.setColor(gframe.config.bgColor);
    offScrGc.fillRect(0,0,getBounds().width,getBounds().height);
    displayTime(offScrGc);
    
    drawBoxGridsTicksLabels(offScrGc,false);  
    SurfaceVertex.setZRange(zmin,zmax);

    boolean customColor = false;

    for (int func = 0; func < gframe.dataSetVector.size(); func++) {
      DataSet2D data = (DataSet2D)gframe.dataSetVector.elementAt(func);

      for (int kl=0; kl<data.getTimeArr().length; kl++) {
        if (data.getTimeValue(kl) == gframe.timeArr[gframe.currentTime]) {

           int dimX = data.timeData[kl].getDimX() - 1;
           int dimY = data.timeData[kl].getDimY() - 1;
           SurfaceVertex[] vertex = data.timeData[kl].vertex;
           if (data.timeData[kl].cArr != null)
             customColor = true;

           for (int k=0; k<vertex.length; k++) {      

              if (vertex[k] == null) return;
              z = vertex[k].z;

              invalid = Float.isNaN(z);
              if (!invalid) {
		if (z >= zmin && z <= zmax) {

                   if (vertex[k] == null) return;
                   projection = vertex[k].projection();
     
                   int val = 0;

                   if (customColor) 
                      val = (int)((data.timeData[kl].cArr[k] - cmin) / (cmax-cmin) * 100f);
                   else
                      val = (int)((z - zmin) / (zmax-zmin) * 100f);

                   if (gframe.colorModel ==GFrame.RGBCOLORMODEL) 
                    offScrGc.setColor(
                      new Color((int)(gframe.rcolors[val] * 255f), 
                         (int)(gframe.gcolors[val] * 255f), 
                         (int)(gframe.bcolors[val] * 255f)));
                   else 
                    offScrGc.setColor(
                      Color.getHSBColor((float)gframe.hcolors[val], 
                                  (float)gframe.scolors[val],
                                  (float)gframe.lcolors[val]));

                   offScrGc.drawOval(projection.x-1, projection.y-1, 2, 2);
		}
	      }
	   }
	
           if (gframe.equallyspaced)
              gframe.monitor.display((gframe.currentTime+1) + " / " + 
         gframe.noOfTime + "   ( " + (dimX+1) +"x"+ (dimY+1) + " )", false);
           else
              gframe.monitor.display((gframe.currentTime+1) + " / " + 
         gframe.noOfTime + " Unequally spaced data", false);
	}
      }
    }

    if (isBoxed) drawBoundingBox();
    image_drawn = true; 
  }


  // Creates wireframe plot 
  private final void plotWireframe() {
    int      i,j,k;
    float    zi,zx;    
    float    z;
    float    lx = 0, ly = 0, lastz = 0;
    Point    lastproj = new Point(0,0);
    boolean  error,lasterror,invalid;
      
    image_drawn = false; 
    projection = new Point(0,0); 

    try {
      zi = gframe.getZMin();
      zx = gframe.getZMax();
      cmin = gframe.getCMin();
      cmax = gframe.getCMax();
      if (zi > zx) throw new NumberFormatException();
      if (zi == zx) 
          zi = zi - 0.0001f;
      if (cmin == cmax) 
          cmin = cmin - 1.0f;
    } catch(NumberFormatException e) {
      gframe.monitor.display("Error in ranges", true); 
      return;
    }

    zmin = zi; zmax = zx; 

    offScrGc.setColor(gframe.config.bgColor);
    offScrGc.fillRect(0,0,getBounds().width,getBounds().height);
    displayTime(offScrGc);
    
    drawBoxGridsTicksLabels(offScrGc,false);  
    offScrGc.setColor(Color.black);

    boolean customColor = false;
 
    SurfaceVertex.setZRange(zmin,zmax);

    for (int func = 0; func < gframe.dataSetVector.size(); func++) {
      DataSet2D data = (DataSet2D)gframe.dataSetVector.elementAt(func);

      for (int kl=0; kl<data.getTimeArr().length; kl++) {
          if (data.getTimeValue(kl) == gframe.timeArr[gframe.currentTime]) {

      int dimX = data.timeData[kl].getDimX() - 1;
      int dimY = data.timeData[kl].getDimY() - 1;
      SurfaceVertex[] vertex = data.timeData[kl].vertex;
      if (data.timeData[kl].cArr != null)
             customColor = true;

      i = 0; j = 0; k = 0; 
      
      // plot - x direction
      while (i <= dimX) {
        lasterror = true;
        while (j <= dimY) {
          z = vertex[k].z;
          invalid = Float.isNaN(z);
          if (!invalid) {
            if (z < zmin) {
              error = true;
              float ratio = (zmin-lastz)/(z-lastz);
              projection = projector.project(ratio*(vertex[k].x-lx)+lx,
                                   ratio*(vertex[k].y-ly)+ly,-10); 
            } else if (z > zmax) {
              error = true;
              float ratio = (zmax-lastz)/(z-lastz);
              projection = projector.project(ratio*(vertex[k].x-lx)+lx,
                                   ratio*(vertex[k].y-ly)+ly,10);
            } else {
              error = false;
              projection = vertex[k].projection();
            }
 
            if (lasterror && (!error) && (j != 0)) {
              if (lastz > zmax) {
                float ratio = (zmax-z)/(lastz-z);
                lastproj = projector.project(
                           ratio*(lx-vertex[k].x)+vertex[k].x,
                           ratio*(ly-vertex[k].y)+vertex[k].y,10);
              } else if (lastz < zmin) {
                float ratio = (zmin-z)/(lastz-z);
                lastproj = projector.project(
                           ratio*(lx-vertex[k].x)+vertex[k].x,
                           ratio*(ly-vertex[k].y)+vertex[k].y,-10);         
              }
            } else invalid = error && lasterror;
          } else error = true;
          if (!invalid && (j != 0)) {
             int val = 0;
             if (customColor) 
                val = (int)((data.timeData[kl].cArr[k] - cmin) / 
                                                 (cmax-cmin) * 100f);
             else if (plot_mode == gframe.WIRECOLORMODE) {
                val = (int)((z - zmin) / (zmax-zmin) * 100f);

                if (gframe.colorModel == GFrame.RGBCOLORMODEL) 
                    offScrGc.setColor(
                      new Color((int)(gframe.rcolors[val] * 255f), 
                                (int)(gframe.gcolors[val] * 255f), 
                                (int)(gframe.bcolors[val] * 255f)));
                else 
                    offScrGc.setColor(
                      Color.getHSBColor((float)gframe.hcolors[val], 
                                        (float)gframe.scolors[val],
                                        (float)gframe.lcolors[val]));
             }

             offScrGc.drawLine(lastproj.x, lastproj.y, 
                                  projection.x, projection.y);
          } 
          lastproj = projection;
          lasterror = error;
          lx = vertex[k].x; 
          ly = vertex[k].y; 
          lastz = z;
          j++; k++;       
        }  
        j = 0; i++; 

      }
    
      // plot - y direction
      i = 0; j = 0; k = 0;

      while (j <= dimY) {
        lasterror = true;
        while (i <= dimX) {

          z = vertex[k].z;
          invalid = Float.isNaN(z);
          if (!invalid) {
            if (z < zmin) {
              error = true;
              float ratio = (zmin-lastz)/(z-lastz);
              projection = projector.project(ratio*(vertex[k].x-lx)+lx,
                                   ratio*(vertex[k].y-ly)+ly,-10); 
            } else if (z > zmax) {
              error = true;
              float ratio = (zmax-lastz)/(z-lastz);
              projection = projector.project(ratio*(vertex[k].x-lx)+lx,
                                   ratio*(vertex[k].y-ly)+ly,10);
            } else {
              error = false;
              projection = vertex[k].projection();
            }

            if (lasterror && (!error) && (i != 0)) {
              if (lastz > zmax) {
                float ratio = (zmax-z)/(lastz-z);
                lastproj = projector.project(
                           ratio*(lx-vertex[k].x)+vertex[k].x,
                           ratio*(ly-vertex[k].y)+vertex[k].y,10);
              } else if (lastz < zmin) {
                float ratio = (zmin-z)/(lastz-z);
                lastproj = projector.project(
                           ratio*(lx-vertex[k].x)+vertex[k].x,
                           ratio*(ly-vertex[k].y)+vertex[k].y,-10);         
              }
            } else invalid = error && lasterror;
          } else error = true;
          if (!invalid && (i != 0)) {
             int val = 0;
             if (customColor) 
                val = (int)((data.timeData[kl].cArr[k] - cmin) / 
                                                 (cmax-cmin) * 100f);
             else if (plot_mode == gframe.WIRECOLORMODE) {
                val = (int)((z - zmin) / (zmax-zmin) * 100f);

                if (gframe.colorModel == GFrame.RGBCOLORMODEL) 
                    offScrGc.setColor(
                      new Color((int)(gframe.rcolors[val] * 255f), 
                                (int)(gframe.gcolors[val] * 255f), 
                                (int)(gframe.bcolors[val] * 255f)));
                else 
                    offScrGc.setColor(
                      Color.getHSBColor((float)gframe.hcolors[val], 
                                        (float)gframe.scolors[val],
                                        (float)gframe.lcolors[val]));
             }

             offScrGc.drawLine(lastproj.x, lastproj.y, 
                                            projection.x, projection.y);
          } 
          lastproj = projection;
          lasterror = error;
          lx = vertex[k].x; 
          ly = vertex[k].y; 
          lastz = z;
          i++; k += dimY+1;       
        }  
        i = 0; k = ++j;
      }
      gframe.monitor.display((gframe.currentTime+1) + " / " + 
         gframe.noOfTime + "   ( " + (dimX+1) +"x"+ (dimY+1) + " )", false);

    } } } // annonnying?

    if (isBoxed) drawBoundingBox();
    image_drawn = true; 
  }


  // display time
  private void displayTime(Graphics g) {
    if (gframe.config.display_time) {
      g.setColor(Color.black);
      FontMetrics fm = g.getFontMetrics();
      g.drawString("Time: " + 
        SVUIC.formatFloat(gframe.timeArr[gframe.currentTime], 3, 2,
                SVUIC.ALGEBRAIC_FORMAT), 5, fm.getAscent()+10);
    }
  }

} // end of class


//////////////////////////////////////////////////////////////////////
class Face {
  public SurfaceVertex[] vertex; 
  public float zdepth;           // z depth of furthest vertex

  Face(SurfaceVertex[] vtx) {
     vertex = new SurfaceVertex[vtx.length];
     System.arraycopy(vtx, 0, vertex, 0, vtx.length);
  }

  void calc_zdepth(SurfaceVertex cop) {
     float a = (vertex[3].y - vertex[1].y) * (vertex[2].z - vertex[1].z) -
                 (vertex[3].z - vertex[1].z) * (vertex[2].y - vertex[1].y);

     float b = (vertex[3].z - vertex[1].z) * (vertex[2].x - vertex[1].x) -
                 (vertex[3].x - vertex[1].x) * (vertex[2].z - vertex[1].z);

     float c = (vertex[3].x - vertex[1].x) * (vertex[2].y - vertex[1].y) -
                 (vertex[3].y - vertex[1].y) * (vertex[2].x - vertex[1].x);

     float d = -1f * (a * vertex[0].x + b * vertex[0].y + c * vertex[0].z);
     zdepth = (a * cop.x + b * cop.y + c* cop.z + d) / 
                               (float)Math.sqrt((double)(a*a + b*b + c*c));
  }
}

///////////////////////////////////
class ContourLineData {
  float[] data;

  ContourLineData(float[] data) {
     this.data = data;
  }

  public float[] getContourLineData() {
    return data;
  }
}
