package sv.kernel;

import java.util.*;

/** 
 * Filter.java
 * abstract super class for filter 
 * All filter implementation will be derived from this class. 
 * @version 1.0     9/11/97
 * @author Byeongseob Ki
 */

public abstract class Filter {

  /**
   * operation over one data set
   */
  public static final int NORMAL  = 1;

  /**
   * operation over multiple data set
   */
  public static final int MULTIPLE = 2;


  /**
   * source data set
   */
  protected Vector sources;

  /**
   * arguments
   * All arguments should be float values.
   */
  protected float args[];

  /**
   * user interface label
   */
  protected String label = "Enter arguments: ";

  // output data set dimension
  private int output_dimension;

  // style of the filter
  private int style;

  // number of arguments
  private int noOfArgs;

  // number of expected data set
  private int expectedSets = -1;

  /**
   * filter constructor. It will be used when the filter does not need
   * an argument. Ex) absolute value caculation filter
   * @param o output data set dimension
   */
  public Filter(int o) {
     output_dimension = o;
     noOfArgs = 0;
     style = NORMAL;

     sources = new Vector();
  }

  /**
   * filter constructor. It will be used when the filter needs argument(s).
   * Ex) x = c*x or f(x) = c*f(x)
   * @param o output data set dimension
   * @param i the number of arguments
   */
  public Filter(int o, int i) {
     output_dimension = o;
     noOfArgs = i;
     style = NORMAL;

     sources = new Vector();
  }

  /**
   * filter constructor. It will be used when the filter needs 
   * multiplae data sets
   * Ex) merge(add) two data set windows
   * @param o output data set dimension
   * @param i the number of arguments
   * @param t the number of expected data set, if you don't know how many,
   *          then just give 999
   */
  public Filter(int o, int i, int t) {
     output_dimension = o;
     noOfArgs = 0;            // IT SHOULD BE i. BUT NOT SUPPORT YET.
     expectedSets = t;
     style = MULTIPLE;

     sources = new Vector();
  }


  /**
   * clear this filter. It will be used by system.
   */
  public final void clear() {
     sources.removeAllElements();
  }

  /**
   * return number of arguments. It will be used by system.
   */
  public final int getNoOfArgs() {
     return noOfArgs;
  }

  /**
   * return number of data set to be need. It will be used by system.
   */
  public final int getExpectDataSets() {
     return expectedSets;
  }

  /**
   * return i-th object of the <i>sources</i> vector.
   * It will be used by system.
   * @param i index of the vector
   */
  public final Object getSourcesAt(int i) {
     return sources.elementAt(i);
  }


  /**
   * return data vector size. It will be used by system.
   */
  public final int getDataVectorSize() {
     return sources.size();
  }

  /**
   * return style of this filter. It will be used by system.
   */
  public final int getStyle() {
     return style;
  }

  /**
   * return out data set dimension. It will be used by system.
   */
  public final int getOutDimension() {
     return output_dimension;
  }

  /**
   * return label for the user interface
   */
  public String getLabel() {
     return label;
  }


  /**
   * assign data set. It will be used by system.
   * @param obj data set object
   */
  public final void assignDataSet(Object obj) {
     sources.addElement(obj);
  }

  /**
   * assign whole data set. It will be used by system.
   * @param w whole data set object
   */
  public final void assignDataSets(Vector w) {
     sources.removeAllElements();
     sources = w;
  }


  /**
   * assign whole arguments. It will be used by system.
   * @param a argument array
   * @return if arguments are successfully assigned, return true
   */
  public final boolean assignArgs(float[] a) {
     if (noOfArgs != a.length)
        return false;
     else 
        args = a;
     return true;
  }


  /**
   * assign only one argument. It will be used by system.
   * @param a argument 
   * @return if arguments are successfully assigned, return true
   */
  public final boolean assignArgs(float a) {
     if (noOfArgs != 1)
        return false;
     else {
        args = new float[1];
        args[0] = a;
     }
     return true;
  }

  /**
   * assign only two arguments. It will be used by system.
   * @param a1 first argument 
   * @param a2 second argument 
   * @return if arguments are successfully assigned, return true
   */
  public final boolean assignArgs(float a1, float a2) {
     if (noOfArgs != 2)
        return false;
     else {
        args = new float[2];
        args[0] = a1;
        args[1] = a2;
     }
     return true;
  }


  /**
   * calculation over source data sets
   * The users need only this function by using instants
   * <i>sources</i> and <i>args</>.
   */
  public abstract Object performFilter();

}





