package sv.kernel;
               
import java.awt.*;
import java.util.*;

import sv.drawer.*;

/** 
 * GCanvas.java
 * Define a canvas to draw the graph
 * @version 1.0    11/1/97
 * @author Byeongseob Ki
 */

// An abstract class is a class that is partially implemented 
// and whose purpose is solely as a design convenience. Abstract class
// are made up of one or more abstract methods, which are methods
// that are declared but left bodiless.

public abstract class GCanvas extends Canvas {

  // double buffering
  protected Graphics offScrGc;
  protected Image offScrImage;

  // printing
  protected boolean printing = false;        // printing flag
  protected int printwidth, printheight;     // print size

  // mouse dragging
  protected boolean isDragged = false;

  // for animator
  public boolean paintDone = true;

  // virtual desktop
  public Rectangle rect;

  // whiteboard manipulation
  public boolean drawingmode = false;
  public DrawHandle dh;
  protected HotSpot hotspot;
  protected ShapeObject cuttedObject;
  protected ShapeObject multiShape;
  protected Vector mPointVec;
  protected PopupMenu wpopup;

  protected MenuItem wcutmi, wcopymi, wpastemi, wlinepmi, wfillmi,
                   wlinecmi, wfillcmi, wtextcmi, wtextfmi;

  /**
   * do whiteboard cut action
   */
  public void doWCut(int index) {
     ShapeObject s = (ShapeObject)dh.drawingVec.elementAt(index);
     Vector hsvec = s.getHotSpots();
     if (hsvec != null)
        dh.hotspotManager.removeElements(hsvec);
     dh.drawingVec.removeElement(s);
     cuttedObject = s;
     repaint();
  }

  /**
   * do whiteboard cut action
   */
  public void doWCopy(int index) {
     ShapeObject s = (ShapeObject)dh.drawingVec.elementAt(index);
     cuttedObject = s;
  }

  /**
   * do whiteboard paste action
   */
  public void doWPaste() {
     dh.manage(cuttedObject);
     repaint();
  }

  /**
   * drawing object creation
   */
  public void doWCreate(String str) {
     dh.createDrawingObject(str);
     repaint();
  }

  /**
   * hotspot object creation
   */
  public void doWHSCreate(int pos) {
     hotspot = dh.hotspotManager.isAt(pos);
  }


  /**
   * hotspot object move
   */
  public void doWHSMove(int x, int y) {
    if (hotspot != null) {
        hotspot.setPoint(new Point(x, y));
        hotspot.notifyObservers();
        hotspot = null;
        repaint();
    }
  }

  /**
   * do fill action
   */
  public void doWFill(int index) {

     ShapeObject s = (ShapeObject)dh.drawingVec.elementAt(index);
     if (s.getType() == ShapeProperties.RECTANGLE) {
        ((RectangleShape)s).changeFillMode();
     } else if (s.getType() == ShapeProperties.POLYGON) {
        ((PolygonShape)s).changeFillMode();
     } else if (s.getType() == ShapeProperties.OVAL) {
        ((OvalShape)s).changeFillMode();
     }
     repaint();
  }


  /**
   * do line color change action
   */
  public void doWLineColor(int index, int c) {
     ShapeObject s = (ShapeObject)dh.drawingVec.elementAt(index);
     if (s.getType() == ShapeProperties.LINE) {
        ((LineShape)s).setLineColor(new Color(c));
     } else if (s.getType() == ShapeProperties.MULTILINE) {
        ((MLineShape)s).setLineColor(new Color(c));
     } else if (s.getType() == ShapeProperties.BEZIER) {
        ((BLineShape)s).setLineColor(new Color(c));
     } else if (s.getType() == ShapeProperties.MOUSE) {
        ((RLineShape)s).setLineColor(new Color(c));
     } else if (s.getType() == ShapeProperties.RECTANGLE) {
        ((RectangleShape)s).setLineColor(new Color(c));
     } else if (s.getType() == ShapeProperties.POLYGON) {
        ((PolygonShape)s).setLineColor(new Color(c));
     } else if (s.getType() == ShapeProperties.OVAL) {
        ((OvalShape)s).setLineColor(new Color(c));
     }
     repaint();
  }

  /**
   * do fill color change action
   */
  public void doWFillColor(int index, int c) {
     ShapeObject s = (ShapeObject)dh.drawingVec.elementAt(index);

     if (s.getType() == ShapeProperties.RECTANGLE) {
        ((RectangleShape)s).setFillColor(new Color(c));
     } else if (s.getType() == ShapeProperties.POLYGON) {
        ((PolygonShape)s).setFillColor(new Color(c));
     } else if (s.getType() == ShapeProperties.OVAL) {
        ((OvalShape)s).setFillColor(new Color(c));
     }
     repaint();
  }

  /**
   * do text color change action
   */
  public void doWTextColor(int index, int c) {
     ShapeObject s = (ShapeObject)dh.drawingVec.elementAt(index);

     if (s.getType() == ShapeProperties.TEXT) {
        ((TextShape)s).setTextColor(new Color(c));
     }
     repaint();
  }

  /**
   * do text font change action
   */
  public void doWTextFont(int index, String name, int style, int size) {
     ShapeObject s = (ShapeObject)dh.drawingVec.elementAt(index);

     if (s.getType() == ShapeProperties.TEXT) {
        ((TextShape)s).setFont(new Font(name, style, size));
     }
     repaint();
  }


  /**
   * do line properties change action
   */
  public void doWPen(int index, int type, int line_wid, int dash_sz) {
     ShapeObject s = (ShapeObject)dh.drawingVec.elementAt(index);

     if (s.getType() == ShapeProperties.LINE) {
        ((LineShape)s).setLineType(type);
        ((LineShape)s).setLineWidth(line_wid);
        ((LineShape)s).setDashSize(dash_sz);

     } else if (s.getType() == ShapeProperties.MULTILINE) {
        ((MLineShape)s).setLineType(type);
        ((MLineShape)s).setLineWidth(line_wid);
        ((MLineShape)s).setDashSize(dash_sz);

     } else if (s.getType() == ShapeProperties.BEZIER) {
        ((BLineShape)s).setLineType(type);
        ((BLineShape)s).setLineWidth(line_wid);
        ((BLineShape)s).setDashSize(dash_sz);

     } else if (s.getType() == ShapeProperties.MOUSE) {
        ((RLineShape)s).setLineType(type);
        ((RLineShape)s).setLineWidth(line_wid);
        ((RLineShape)s).setDashSize(dash_sz);
     }

     repaint();
  }

}
