package sv.kernel;
               
import java.awt.*;
import java.util.*;
import java.awt.event.*;
import java.io.*;

import sv.ui.SimpleMonitor;
import sv.util.NewtonMethod;

/** 
 * GFrame.java
 * Define a frame to draw the graph
 * @version 1.0    8/1/97
 * @author Byeongseob Ki
 */

// An abstract class is a class that is partially implemented 
// and whose purpose is solely as a design convenience. Abstract class
// are made up of one or more abstract methods, which are methods
// that are declared but left bodiless.

public abstract class GFrame extends Frame {

  public static final int RGBCOLORMODEL = 1;
  public static final int HSBCOLORMODEL = 2;

  /**
   * id number of the visualization window
   */
  protected long id;

  /**
   * dimension of the data set
   */
  protected int dimension_;

  /**
   * title of this visualization window
   */
  protected String frameTitle;

  /**
   * print properties
   */
  protected Properties printerprops; 

  /** 
   * to display the information of the graph and error message
   */
  public SimpleMonitor monitor;

  /**
   * DataSet vector
   */
  public Vector dataSetVector;

  /**
   * time data array for all dataset attached this frame
   */
  public float[] timeArr;

  /**
   * the size of time data array for all dataset attached this frame
   */
  public int noOfTime = 1;

  /**
   * indicate the current time
   */
  public int currentTime = 0;

  /**
   * if set true, use global min and max to plot data set
   */
  public boolean globalFlag = true;      

  /**
   * if set true, stealth mode is on for collaboration
   */
  public boolean stealthMode = false;

  /**
   * buffer size
   */
  protected int bufferSize = Integer.MAX_VALUE;

  // drawing handle vector
  protected Vector DHV;

  // for color map
  public int colorModel  = 1;
  public NewtonMethod rspline, gspline, bspline, 
                      hspline, sspline, lspline;    

  /**
   * constructor(to create peer frame for split command)
   * @param title title of this class
   */
  public GFrame(String title) {
     super(title);
     frameTitle = title;
     id = System.currentTimeMillis();
     dataSetVector = new Vector();

     printerprops = new Properties();
  }   

  /**
   * return id of this visualization window
   */
  public long getID() {
    return id;
  }

  /**
   * set id for this visualization window
   * @param frameID id for this visualization window
   */
  public void setID(long frameID) {
    id = frameID;
  }


  /**
   * return frameTitle
   */
  public String getFrameTitle() {
     return frameTitle;
  }


  /**
   * set title of this frame
   * @param title title of this frame
   */
  public void setFrameTitle(String title) {
     frameTitle = title;
     setTitle(title);
  }

  /**
   * return dimension of this frame
   */
  public int getRank() {
     return dimension_;
  }

  /**
   * return buffer size of this frame
   */
  public int getBufferSize() {
     return bufferSize;
  }

  /**
   * set buffer size of this frame
   * @param sz buffer size
   */
  public void setBufferSize(int sz) {
     bufferSize = sz;
  }

  /**
   * change color map info
   */
  public abstract void changeColorMap(String[] args);

  /**
   * return color info
   */
  public String[] getColorInfo() {
     String[] str = new String[2];
     str[0] = String.valueOf(colorModel);
     str[1] = "";

     for (int i=0; i<101; i++) {
        str[1] += Float.toString(rcolors[i]) + ";";
        str[1] += Float.toString(gcolors[i]) + ";";
        str[1] += Float.toString(bcolors[i]) + ";";
        str[1] += Float.toString(hcolors[i]) + ";";
        str[1] += Float.toString(scolors[i]) + ";";
        str[1] += Float.toString(lcolors[i]) + ";";
     }
     return str;
  }


  /**
   * set color map
   */
  public void setColorMap(String[] str) {
     colorModel = Integer.valueOf(str[0]).intValue();
     StringTokenizer st = new StringTokenizer(str[1], ";");
     for (int i=0; i<101; i++) {
        rcolors[i] = Float.valueOf(st.nextToken()).floatValue();
        gcolors[i] = Float.valueOf(st.nextToken()).floatValue();
        bcolors[i] = Float.valueOf(st.nextToken()).floatValue();
        hcolors[i] = Float.valueOf(st.nextToken()).floatValue();
        scolors[i] = Float.valueOf(st.nextToken()).floatValue();
        lcolors[i] = Float.valueOf(st.nextToken()).floatValue();
     }
  }


  /**
   * update timeArray and noOfTime values
   */
  protected void updateTimeArr() {
     Enumeration e;
     float[] tmpArr1 = null, tmpArr2 = null, wholeArr;
     int tmpsz1 = 0, tmpsz2 = 0, index, index1, index2;
     boolean needSort = false;
     boolean done = false;

     for (e=dataSetVector.elements(); e.hasMoreElements();) {
        DataSet data = (DataSet)e.nextElement();

        if (needSort) {

           // copy next DataSet info first
           tmpsz2 = data.getnoOfTime();
           tmpArr2 = new float[tmpsz2];
           tmpArr2 = data.getTimeArr();
          
           wholeArr = new float[tmpsz1 + tmpsz2];

           index = 0; index1 = 0; index2 = 0;
           done = false;
           while(!done) {
              if (tmpArr1[index1] < tmpArr2[index2]) {
                 wholeArr[index] = tmpArr1[index1];
                 index1++;
              } else if (tmpArr1[index1] > tmpArr2[index2]) {
                 wholeArr[index] = tmpArr2[index2];
                 index2++;
              } else {
                 wholeArr[index] = tmpArr1[index1];
                 index1++; index2++;
              }
              index++;

              if (index1 == tmpsz1) {
                 for (int i=index2; i<tmpsz2; i++) {
                    wholeArr[index] = tmpArr2[i];
                    index++;
                 }
                 done = true;
              } else if (index2 == tmpsz2) {
                 for (int i=index1; i<tmpsz1; i++) {
                    wholeArr[index] = tmpArr1[i];
                    index++;
                 }
                 done = true;
              }
           }
           tmpArr1 = new float[index];
           for (int i=0; i<index; i++)
              tmpArr1[i] = wholeArr[i];
           tmpsz1 = index;      

        } else {     // first element of the dataset
           tmpArr1 = data.getTimeArr();
           tmpsz1 = data.getnoOfTime();
           needSort = true;
        }
     }   
     timeArr = new float[tmpsz1];
     timeArr = tmpArr1;
     noOfTime = tmpsz1;
  }



  /**
   * read the color map file 
   */
  protected boolean readColorMap(String fname) {

     try {
        BufferedReader reader = new BufferedReader(new FileReader(fname));

        int nControls;
        float[] xcontrols, ycontrols;
        String input, type;
        StringTokenizer st;

        for (int tt=0; tt<6; tt++) {
           input = reader.readLine();
           st = new StringTokenizer(input);
           type = st.nextToken().toLowerCase();

           if (type.indexOf("r") >= 0) {
              nControls = Integer.valueOf(st.nextToken()).intValue();

              xcontrols = new float[nControls];
              ycontrols = new float[nControls];

              for (int i=0; i<nControls; i++) {
                input = reader.readLine();
		st = new StringTokenizer(input);

                xcontrols[i] = Float.valueOf(st.nextToken()).floatValue();
                ycontrols[i] = Float.valueOf(st.nextToken()).floatValue();
              }
   
              rspline = new NewtonMethod(xcontrols, ycontrols);
              
           } else if (type.indexOf("g") >= 0) {
              nControls = Integer.valueOf(st.nextToken()).intValue();

              xcontrols = new float[nControls];
              ycontrols = new float[nControls];

              for (int i=0; i<nControls; i++) {
                input = reader.readLine();
		st = new StringTokenizer(input);

                xcontrols[i] = Float.valueOf(st.nextToken()).floatValue();
                ycontrols[i] = Float.valueOf(st.nextToken()).floatValue();
              }
   
              gspline = new NewtonMethod(xcontrols, ycontrols);

           } else if (type.indexOf("b") >= 0) {
              nControls = Integer.valueOf(st.nextToken()).intValue();

              xcontrols = new float[nControls];
              ycontrols = new float[nControls];

              for (int i=0; i<nControls; i++) {
                input = reader.readLine();
		st = new StringTokenizer(input);

                xcontrols[i] = Float.valueOf(st.nextToken()).floatValue();
                ycontrols[i] = Float.valueOf(st.nextToken()).floatValue();
              }
   
              bspline = new NewtonMethod(xcontrols, ycontrols);

           } else if (type.indexOf("h") >= 0) {
              nControls = Integer.valueOf(st.nextToken()).intValue();

              xcontrols = new float[nControls];
              ycontrols = new float[nControls];

              for (int i=0; i<nControls; i++) {
                input = reader.readLine();
		st = new StringTokenizer(input);

                xcontrols[i] = Float.valueOf(st.nextToken()).floatValue();
                ycontrols[i] = Float.valueOf(st.nextToken()).floatValue();
              }
   
              hspline = new NewtonMethod(xcontrols, ycontrols);

           } else if (type.indexOf("s") >= 0) {
              nControls = Integer.valueOf(st.nextToken()).intValue();

              xcontrols = new float[nControls];
              ycontrols = new float[nControls];

              for (int i=0; i<nControls; i++) {
                input = reader.readLine();
		st = new StringTokenizer(input);

                xcontrols[i] = Float.valueOf(st.nextToken()).floatValue();
                ycontrols[i] = Float.valueOf(st.nextToken()).floatValue();
              }
   
              sspline = new NewtonMethod(xcontrols, ycontrols);

           } else if (type.indexOf("l") >= 0) {
              nControls = Integer.valueOf(st.nextToken()).intValue();

              xcontrols = new float[nControls];
              ycontrols = new float[nControls];

              for (int i=0; i<nControls; i++) {
                input = reader.readLine();
		st = new StringTokenizer(input);

                xcontrols[i] = Float.valueOf(st.nextToken()).floatValue();
                ycontrols[i] = Float.valueOf(st.nextToken()).floatValue();
              }
   
              lspline = new NewtonMethod(xcontrols, ycontrols);

           } else {
             System.out.println("Can't read colormap file...");
             reader.close();
             return false;
           }
        }
        reader.close();
        return true;

     } catch (Exception e) {
        e.printStackTrace();
        System.out.println("Error while reading the colormap file ...");
        System.out.println("This may cause undesirable color effects ...");
        return false;
     }
  }



public float rcolors[] = { 
0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.02f, 0.03f, 
0.05f, 0.07f, 0.10f, 0.12f, 0.15f, 0.18f, 0.21f, 0.24f, 0.27f, 0.30f, 
0.33f, 0.37f, 0.40f, 0.43f, 0.47f, 0.50f, 0.53f, 0.57f, 0.60f, 0.63f, 
0.66f, 0.69f, 0.72f, 0.75f, 0.78f, 0.80f, 0.83f, 0.85f, 0.87f, 0.89f, 
0.91f, 0.93f, 0.94f, 0.95f, 0.97f, 0.98f, 0.99f, 0.99f, 1.00f, 1.00f,
1.00f, 1.00f, 1.00f, 0.99f, 0.99f, 0.98f, 0.97f, 0.95f, 0.94f, 0.93f, 
0.91f, 0.89f, 0.87f, 0.85f, 0.83f, 0.80f, 0.78f, 0.75f, 0.72f, 0.70f, 
0.66f, 0.63f, 0.60f, 0.57f, 0.53f, 0.50f, 0.47f, 0.43f, 0.40f, 0.37f, 
0.33f, 0.30f, 0.27f, 0.24f, 0.21f, 0.18f, 0.15f, 0.12f, 0.10f, 0.07f, 
0.05f, 0.03f, 0.02f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 
0.0f};

public float gcolors[] = { 
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f };

public float bcolors[] = { 
1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 0.98f, 0.97f, 
0.95f, 0.93f, 0.90f, 0.88f, 0.85f, 0.82f, 0.79f, 0.76f, 0.73f, 0.70f, 
0.67f, 0.63f, 0.60f, 0.57f, 0.53f, 0.50f, 0.47f, 0.43f, 0.40f, 0.37f, 
0.34f, 0.31f, 0.28f, 0.25f, 0.22f, 0.20f, 0.17f, 0.15f, 0.13f, 0.11f, 
0.09f, 0.07f, 0.06f, 0.05f, 0.03f, 0.02f, 0.01f, 0.01f, 0.00f, 0.00f,
0.00f, 0.00f, 0.00f, 0.01f, 0.01f, 0.02f, 0.03f, 0.05f, 0.06f, 0.07f, 
0.09f, 0.11f, 0.13f, 0.15f, 0.17f, 0.20f, 0.22f, 0.25f, 0.28f, 0.31f, 
0.34f, 0.37f, 0.40f, 0.43f, 0.47f, 0.50f, 0.53f, 0.57f, 0.60f, 0.63f, 
0.67f, 0.70f, 0.73f, 0.76f, 0.79f, 0.82f, 0.85f, 0.88f, 0.90f, 0.93f, 
0.95f, 0.97f, 0.98f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f,
1.00f }; 


public float hcolors[] = { 
0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.02f, 0.03f, 
0.05f, 0.07f, 0.10f, 0.12f, 0.15f, 0.18f, 0.21f, 0.24f, 0.27f, 0.30f, 
0.33f, 0.37f, 0.40f, 0.43f, 0.47f, 0.50f, 0.53f, 0.57f, 0.60f, 0.63f, 
0.66f, 0.69f, 0.72f, 0.75f, 0.78f, 0.80f, 0.83f, 0.85f, 0.87f, 0.89f, 
0.91f, 0.93f, 0.94f, 0.95f, 0.97f, 0.98f, 0.99f, 0.99f, 1.00f, 1.00f,
1.00f, 1.00f, 1.00f, 0.99f, 0.99f, 0.98f, 0.97f, 0.95f, 0.94f, 0.93f, 
0.91f, 0.89f, 0.87f, 0.85f, 0.83f, 0.80f, 0.78f, 0.75f, 0.72f, 0.70f, 
0.66f, 0.63f, 0.60f, 0.57f, 0.53f, 0.50f, 0.47f, 0.43f, 0.40f, 0.37f, 
0.33f, 0.30f, 0.27f, 0.24f, 0.21f, 0.18f, 0.15f, 0.12f, 0.10f, 0.07f, 
0.05f, 0.03f, 0.02f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 0.00f, 
0.0f};

public float scolors[] = { 
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f,
0.5f };

public float lcolors[] = { 
1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 0.98f, 0.97f, 
0.95f, 0.93f, 0.90f, 0.88f, 0.85f, 0.82f, 0.79f, 0.76f, 0.73f, 0.70f, 
0.67f, 0.63f, 0.60f, 0.57f, 0.53f, 0.50f, 0.47f, 0.43f, 0.40f, 0.37f, 
0.34f, 0.31f, 0.28f, 0.25f, 0.22f, 0.20f, 0.17f, 0.15f, 0.13f, 0.11f, 
0.09f, 0.07f, 0.06f, 0.05f, 0.03f, 0.02f, 0.01f, 0.01f, 0.00f, 0.00f,
0.00f, 0.00f, 0.00f, 0.01f, 0.01f, 0.02f, 0.03f, 0.05f, 0.06f, 0.07f, 
0.09f, 0.11f, 0.13f, 0.15f, 0.17f, 0.20f, 0.22f, 0.25f, 0.28f, 0.31f, 
0.34f, 0.37f, 0.40f, 0.43f, 0.47f, 0.50f, 0.53f, 0.57f, 0.60f, 0.63f, 
0.67f, 0.70f, 0.73f, 0.76f, 0.79f, 0.82f, 0.85f, 0.88f, 0.90f, 0.93f, 
0.95f, 0.97f, 0.98f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f, 1.00f,
1.00f }; 


}
