package sv.kernel;

import java.awt.*;
import java.io.*;
import java.util.*;

import sv.util.*;

/** 
 * TimeData2D.java
 * Define point data in each time
 * @version 1.0    10/1/97
 * @author Byeongseob Ki
 */

public class TimeData2D extends TimeData  {

  private float xmin, ymin, zmin, xmax, ymax, zmax, cmin, cmax;
  private float[] xArr = null, yArr = null, zArr = null;
  public float[] cArr = null;

  public SurfaceVertex[] vertex = null;
   
  private int dimX;
  private int dimY;

  private TimeDataVec vectorSet = null;

  private boolean bb = false;
  private boolean equallyspaced = false;
  

  /**
   * constructor for 2d dataset
   * @param time time value
   * @param noOfPoints the number of points of this data
   * @param xArr x value array
   * @param yArr y value array
   * @param zArr y value array
   */
  public TimeData2D(float time, int noOfPoints,
                             float[] xArr, float[] yArr, float[] zArr) {
     this.time = time;
     this.noOfPoints = noOfPoints;

     this.xArr = xArr;
     this.yArr = yArr;
     this.zArr = zArr;

     initialize(false);
  }


  /**
   * constructor for 2d dataset (called by java_bbser2d)
   * @param time time value
   * @param noOfPoints the number of points of this data
   * @param xmin xmin value
   * @param xmax x max
   * @param ymin ymin value
   * @param ymax y max
   * @param zArr z value array
   */
  public TimeData2D(float time,
         float xmin, float xmax, int nx, 
         float ymin, float ymax, int ny, float[] zArr) {
     this.time = time;
     this.noOfPoints = noOfPoints;

     this.zArr = zArr;
     bb = true;
     equallyspaced = true;

     this.xmin = (float)xmin;
     this.ymin = (float)ymin;
     zmin = 1.0e+10f;
     this.xmax = (float)xmax;
     this.ymax = (float)ymax;
     zmax = 1.0e-10f;

     noOfPoints = nx * ny;

     for (int i=0; i<noOfPoints; i++) {
        if (zmin > zArr[i]) zmin = zArr[i];
        if (zmax < zArr[i]) zmax = zArr[i];
     }

     dimX = nx;
     dimY = ny;
  }


  /**
   * constructor for 2d dataset (called by java_cbbser2d)
   * @param time time value
   * @param noOfPoints the number of points of this data
   * @param xmin xmin value
   * @param xmax x max
   * @param ymin ymin value
   * @param ymax y max
   * @param zArr z value array
   * @param color color array
   */
  public TimeData2D(float time,
         float xmin, float xmax, int nx, 
         float ymin, float ymax, int ny, float[] zArr, float[] color) {

     this(time, xmin, xmax, nx, ymin, ymax, ny, zArr);
     cArr = color; 

     cmin = 1.0e+10f;
     cmax = 1.0e-10f;

     for (int i=0; i<noOfPoints; i++) {
        if (cmin > cArr[i]) cmin = cArr[i];
        if (cmax < cArr[i]) cmax = cArr[i];
     }

  }

  /**
   * constructor for 2d dataset
   * @param time time value
   * @param xArr x array
   * @param nx number of x points
   * @param yArr y array
   * @param ny number of y points
   * @param zArr z value array
   */
  public TimeData2D(float time, 
         float[] xArr, int nx, float[] yArr, int ny, float[] zArr) {
     this.time = time;
     this.noOfPoints = nx * ny;

     equallyspaced = true;

     this.zArr = zArr;
     this.xArr = new float[zArr.length];
     this.yArr = new float[zArr.length];

     for (int i=0; i<nx; i++) {
        int index = i*ny;
        for (int j=0; j<ny; j++) {
           this.xArr[index+j] = xArr[i];
           this.yArr[index+j] = yArr[j];
        }
     }

     this.xmin = xArr[0];
     this.ymin = yArr[0];
     zmin = 1.0e+10f;
     this.xmax = xArr[nx-1];
     this.ymax = yArr[ny-1];
     zmax = 1.0e-10f;

     for (int i=0; i<noOfPoints; i++) {
        if (zmin > zArr[i]) zmin = zArr[i];
        if (zmax < zArr[i]) zmax = zArr[i];
     }

     dimX = nx;
     dimY = ny;
  }


  /** 
   * sending data by using DataOutputStream
   */
  public void sendingData(DataOutputStream out) {

    try {

       // send time
       out.writeFloat(time);

       // send number of points in given time
       out.writeInt(dimX);
       out.writeInt(dimY);

       // send cArr
       if (cArr == null) {
          out.writeByte(0);
       } else {
          out.writeByte(1);
          for (int i=0; i<noOfPoints; i++) 
             out.writeFloat(cArr[i]);
       }

       // send vector set
       if (vectorSet == null) {
          out.writeByte(0);
       } else {
          out.writeByte(1);
          vectorSet.sendingData(out);
       }

       // send reserved for future use (2 items)
       out.writeInt(0);
       out.writeInt(0);

       // send x, y, and z arrays
       float[] zArr = getzArr();
       if (bb) {
         out.writeByte(1);     // bb is true... send boundbox first
         out.writeFloat(xmin);
         out.writeFloat(xmax);
         out.writeFloat(ymin);
         out.writeFloat(ymax);

         for (int i=0; i<noOfPoints; i++) 
           out.writeFloat(zArr[i]);

       } else {
         float[] xArr = getxArr();
         float[] yArr = getyArr();
         out.writeByte(0);      // bb is false

         for (int i=0; i<noOfPoints; i++) {
            out.writeFloat(xArr[i]);
            out.writeFloat(yArr[i]);
            out.writeFloat(zArr[i]);
         }
       }
    } catch (Exception e) {
       System.out.println("Data sending error(TimeData2D).");
e.printStackTrace();
    } 
  }


  private void initialize(boolean b) {

     boolean clearup = false;

     xmin = 1.0e10f;
     ymin = 1.0e10f;
     zmin = 1.0e10f;
     xmax = -1.0e10f;
     ymax = -1.0e10f;
     zmax = -1.0e10f;

     Vector xvec = new Vector();
     Vector yvec = new Vector();
     Float xval, yval;

     if (zArr == null) {
       if (vertex != null) {
          clearup = true;
          zArr = new float[vertex.length];
          for (int i=0; i<vertex.length; i++) 
             zArr[i] = vertex[i].z;

       } else {
          System.out.println("2d initializing error ...");
          return;
       }
     }

     for (int i=0; i<noOfPoints; i++) {
        if (xmin > xArr[i]) xmin = xArr[i];
        if (ymin > yArr[i]) ymin = yArr[i];
        if (zmin > zArr[i]) zmin = zArr[i];
        if (xmax < xArr[i]) xmax = xArr[i];
        if (ymax < yArr[i]) ymax = yArr[i];
        if (zmax < zArr[i]) zmax = zArr[i];

        xval = new Float(xArr[i]);
        yval = new Float(yArr[i]);
        if (!xvec.contains(xval)) 
           xvec.addElement(xval);
        if (!yvec.contains(yval))
           yvec.addElement(yval);
     }

     dimX = xvec.size();
     dimY = yvec.size();

     if (noOfPoints == (dimX*dimY))
        equallyspaced = true;
     else
        equallyspaced = false;
        

     xvec.removeAllElements();
     yvec.removeAllElements();

     if (equallyspaced) {
        xsort(xArr, yArr, zArr, 0, noOfPoints-1);
     
        int[] posArr = new int[dimX+1];
        int idx = 1;
        posArr[0] = 0;

        for (int i=1; i<noOfPoints; i++) 
          if (xArr[i-1] != xArr[i]) {
             posArr[idx] = i;
             idx++;
          } 
        posArr[dimX] = noOfPoints;
             
        for (int i=0; i<dimX; i++) 
           ysort(xArr, yArr, zArr, posArr[i], posArr[i+1]-1);
     }


     if (b) {      // if updated

        float[] savedxArr = new float[noOfPoints];
        float[] savedyArr = new float[noOfPoints];
        float[] savedzArr = new float[noOfPoints];

        int index = 0;
        savedxArr[0] = xArr[0];
        savedyArr[0] = yArr[0];
        savedzArr[0] = zArr[0];
        index++;

        for (int i=1; i<noOfPoints; i++) {
          if ((xArr[i-1] == xArr[i]) && (yArr[i-1] == yArr[i])) {
               // skip                 
          } else {
            savedxArr[index] = xArr[i];
            savedyArr[index] = yArr[i];
            savedzArr[index] = zArr[i];
            index++;
          }
        }
        xArr = new float[index];
        yArr = new float[index];
        zArr = new float[index];

        System.arraycopy(savedxArr, 0, xArr, 0, index);
        System.arraycopy(savedyArr, 0, yArr, 0, index);
        System.arraycopy(savedzArr, 0, zArr, 0, index);

        noOfPoints = index;
     }   

     if (clearup)
       zArr = null;
  }


  public void setVectorSet(TimeDataVec vs) {
     vectorSet = vs;
  }

  public TimeDataVec getVectorSet() {
     return vectorSet;
  }
 
  public void makeVertex(boolean b, float[] minmaxArr) {
     float xfactor, yfactor, xval, yval;

     if (zArr == null)     
        zArr = getzArr();

     if (bb) {
       xArr = new float[zArr.length];
       yArr = new float[zArr.length];

       float xinc = (float)(xmax - xmin) / (float)(dimX-1);
       float yinc = (float)(ymax - ymin) / (float)(dimY-1);

       for (int i=0; i<dimX; i++) {
          int index = i*dimY;
          for (int j=0; j<dimY; j++) {
             xArr[index+j] = (float)(xmin + xinc * i);  
             yArr[index+j] = (float)(ymin + yinc * j);  
          }
        }
     }

     try {
       if (vertex != null) {
          for (int i=0; i<vertex.length; i++)
             vertex[i] = null;
          vertex = null;
       }

       vertex = new SurfaceVertex[noOfPoints];
     } catch(OutOfMemoryError e) {
       System.out.println("Not enough memory");
     } catch(Exception e) {
       System.out.println("Error: " + e.toString());
     }

     if (vertex == null) return;

     if (b) {
        xfactor = 20/(minmaxArr[1] - minmaxArr[0]);
        yfactor = 20/(minmaxArr[3] - minmaxArr[2]);

        for (int i=0; i<noOfPoints; i++) {
           xval = (xArr[i]-minmaxArr[0])*xfactor - 10.0f;
           yval = (yArr[i]-minmaxArr[2])*yfactor - 10.0f;
           vertex[i] = new SurfaceVertex(xval, yval, zArr[i]);
        }

     } else {
        xfactor = 20/(xmax - xmin);
        yfactor = 20/(ymax - ymin);

        for (int i=0; i<noOfPoints; i++) {
           xval = (xArr[i]-xmin)*xfactor - 10.0f;
           yval = (yArr[i]-ymin)*yfactor - 10.0f;
           vertex[i] = new SurfaceVertex(xval, yval, zArr[i]);
        }
     }    

     if (bb) {
        xArr = null;
        yArr = null;
     }
     zArr = null;
  }
  
  /**
   * normalize vector set
   */
  public void normalizeVectorSet(boolean b, float[] minmaxArr) {
    /***********************************************************
     * SHOULD BE CHECK LATER
     if (vectorSet == null)
        return;

     float xfactor, yfactor;
     int size = vectorSet.xArr.length;

     if (b) {
        xfactor = 20/(minmaxArr[1] - minmaxArr[0]);
        yfactor = 20/(minmaxArr[3] - minmaxArr[2]);

        for (int i=0; i<size; i++) {
           vectorSet.xArr[i] = 
                    (vectorSet.xArr[i]-minmaxArr[0])*xfactor - 10.0f;
           vectorSet.yArr[i] = 
                    (vectorSet.yArr[i]-minmaxArr[2])*yfactor - 10.0f;
        }

     } else {
        xfactor = 20/(xmax - xmin);
        yfactor = 20/(ymax - ymin);

        for (int i=0; i<size; i++) {
           vectorSet.xArr[i] = 
                    (vectorSet.xArr[i]-xmin)*xfactor - 10.0f;
           vectorSet.yArr[i] =  
                    (vectorSet.yArr[i]-ymin)*yfactor - 10.0f;
        }
     }
     */   
  }


  /**
   * update data 
   * @param data TimeData2D to be appended or updated
   */
  public void updateData(TimeData2D data) {

    // IF data IS ALSO bb, THEN WE CAN REDUCE MEMORY SPACE MORE!

     if (zArr == null) {
       if (vertex != null) {
          zArr = new float[vertex.length];
          for (int i=0; i<vertex.length; i++) 
             zArr[i] = vertex[i].z;

       } else {
          System.out.println("2d initializing error ...");
          return;
       }
     }

     if (bb) {
       xArr = new float[zArr.length];
       yArr = new float[zArr.length];

       float xinc = (float)(xmax - xmin) / (float)dimX;
       float yinc = (float)(ymax - ymin) / (float)dimY;

       for (int i=0; i<dimX; i++) {
          int index = i*dimY;
          for (int j=0; j<dimY; j++) {
             xArr[index+j] = (float)(xmin + xinc * i);  
             yArr[index+j] = (float)(ymin + yinc * j);  
          }
        }
     }

     bb = false;

     float[] newxArr = data.getxArr();
     float[] newyArr = data.getyArr();
     float[] newzArr = data.getzArr();
     
     float[] savedxArr = new float[noOfPoints];
     float[] savedyArr = new float[noOfPoints];
     float[] savedzArr = new float[noOfPoints];

     System.arraycopy(xArr, 0, savedxArr, 0, noOfPoints);
     System.arraycopy(yArr, 0, savedyArr, 0, noOfPoints);
     System.arraycopy(zArr, 0, savedzArr, 0, noOfPoints);

     int saved = noOfPoints;
     noOfPoints = noOfPoints + newxArr.length;

     xArr = new float[noOfPoints];
     yArr = new float[noOfPoints];
     zArr = new float[noOfPoints];

     System.arraycopy(savedxArr, 0, xArr, 0, saved);
     System.arraycopy(savedyArr, 0, yArr, 0, saved);
     System.arraycopy(savedzArr, 0, zArr, 0, saved);

     int index = saved;
     for (int i=0; i<data.getnoOfPoints(); i++) {
        xArr[index] = newxArr[i];
        yArr[index] = newyArr[i];
        zArr[index] = newzArr[i];
        index++;
     }

     if ((noOfPoints != yArr.length) || (noOfPoints != xArr.length)) { 
         System.out.println("Update TimeData failed ......");
     }

     data = null;            // garbage collection
     initialize(true);
  }


  /**
   * return noOfPoints
   */
  public int getnoOfPoints() {
     return noOfPoints;
  }


  /**
   * return xArr
   */
  public float[] getxArr() {
     int length;

     if (zArr != null) length = zArr.length;
     else if (vertex != null) length = vertex.length;
     else length = dimX * dimY;

     if (bb) {
       float[] xxArr = new float[length];
       float xinc = (float)(xmax - xmin) / (float)(dimX-1);

       for (int i=0; i<dimX; i++) {
          int index = i*dimY;
          for (int j=0; j<dimY; j++) 
             xxArr[index+j] = (float)(xmin + xinc * i);  
        }
        return xxArr;
     } else {
        float[] xxArr = new float[length];
        System.arraycopy(xArr, 0, xxArr, 0, length);
        return xxArr;
     }
  }

  /**
   * return yArr
   */
  public float[] getyArr() {
     int length;

     if (zArr != null) length = zArr.length;
     else if (vertex != null) length = vertex.length;
     else length = dimX * dimY;

     if (bb) {
       float[] yyArr = new float[length];
       float yinc = (float)(ymax - ymin) / (float)(dimY-1);

       for (int i=0; i<dimX; i++) {
          int index = i*dimY;
          for (int j=0; j<dimY; j++) 
             yyArr[index+j] = (float)(ymin + yinc * j);  
        }
        return yyArr;
     } else {
        float[] yyArr = new float[length];
        System.arraycopy(yArr, 0, yyArr, 0, length);
        return yyArr;
     }
  }


  /**
   * return zArr
   */
  public float[] getzArr() {
     if (zArr == null) {
       float[] zzArr = null;
       if (vertex != null) {
          zzArr = new float[vertex.length];
          for (int i=0; i<vertex.length; i++) 
             zzArr[i] = vertex[i].z;

       } else {
          System.out.println("2d initializing error ...");
          return null;
       }
       return zzArr;
     } else {
        int length = zArr.length;
        float[] zzArr = new float[length];
        System.arraycopy(zArr, 0, zzArr, 0, length);
        return zzArr;
     }
  }


  public float getxmin() {
     return xmin;
  }

  public float getxmax() {
     return xmax;
  }


  public float getymin() {
     return ymin;
  }

  public float getymax() {
     return ymax;
  }


  public float getzmin() {
     return zmin;
  }

  public float getzmax() {
     return zmax;
  }

  public float getcmin() {
     return cmin;
  }

  public float getcmax() {
     return cmax;
  }


  public int getDimX() {
     return dimX;
  }	

  public int getDimY() {
     return dimY;
  }	


  public boolean isBounded() {
     return bb;
  }

  public boolean isCustomColor() {
     if (cArr == null)
       return false;
     else
       return true;
  }


  public boolean isEquallySpaced() {
     return equallyspaced;
  }


  private void xsort(float[] arr1, float[] arr2, float[] arr3, int lo0, int hi0) {
     int lo = lo0;
     int hi = hi0;
     float mid;

     if (hi0 > lo0) {

        mid = arr1[(lo0+hi0)/2];

        while (lo <= hi) {
           while ((lo < hi0) && (arr1[lo] < mid))
	       ++lo;

           while (( hi > lo0) && (arr1[hi] > mid))
	      --hi;

           // if the indexes have not crossed, swap
           if (lo <= hi)  {
              float S = arr1[lo]; 
              float T = arr2[lo];
              float U = arr3[lo];
              arr1[lo] = arr1[hi];
              arr2[lo] = arr2[hi];
              arr3[lo] = arr3[hi];
              arr1[hi] = S;
              arr2[hi] = T;
              arr3[hi] = U;
              ++lo;
              --hi;
           }
        }


        if (lo0 < hi)
           xsort(arr1, arr2, arr3, lo0, hi);

        if (lo < hi0)
           xsort(arr1, arr2, arr3, lo, hi0);

      }
  }


  private void ysort(float[] arr1, float[] arr2, float[] arr3, int lo0, int hi0) {
     int lo = lo0;
     int hi = hi0;
     float mid;

     if (hi0 > lo0) {

        mid = arr2[(lo0+hi0)/2];

        while (lo <= hi) {
           while ((lo < hi0) && (arr2[lo] < mid))
	       ++lo;

           while (( hi > lo0) && (arr2[hi] > mid))
	      --hi;

           // if the indexes have not crossed, swap
           if (lo <= hi)  {
              float S = arr1[lo]; 
              float T = arr2[lo];
              float U = arr3[lo];
              arr1[lo] = arr1[hi];
              arr2[lo] = arr2[hi];
              arr3[lo] = arr3[hi];
              arr1[hi] = S;
              arr2[hi] = T;
              arr3[hi] = U;
              ++lo;
              --hi;
           }
        }


        if (lo0 < hi)
           ysort(arr1, arr2, arr3, lo0, hi);

        if (lo < hi0)
           ysort(arr1, arr2, arr3, lo, hi0);

      }
  }


} // end of class
