package sv.kernel;

import java.awt.*;
import java.io.*;
import java.util.*;

/** 
 * TimeDataCon.java
 * Define Connectivity data set
 * @version 1.0   11/15/97
 * @author Byeongseob Ki
 */

public class TimeDataCon extends TDPlane {

  // arrays
  private int[] connection;     // connection list
  private float[][] colors;

  private boolean initialized = false;

  public TimeDataCon(float time, float[] xArr, float[] yArr,
                                float[] cArr, int[] connection) {
     super(time, xArr.length, xArr, yArr, cArr);
     dataType = DSPlane.CONNECTIVITY;

     this.connection = connection;
  }


  /** 
   * sending data by using DataOutputStream
   */
  public void sendingData(DataOutputStream out) {

    try {
       // send time
       out.writeFloat(time);

       // send number of points in given time
       out.writeInt(noOfPoints);

       // send reserved for future use (2 items)
       out.writeInt(0);
       out.writeInt(0);
       
       // send x and y arrays
       if (bb) {
         out.writeByte(1);     // bb is true... send boundbox first
         out.writeFloat(xmin);
         out.writeFloat(xmax);

         for (int i=0; i<noOfPoints; i++) {
           out.writeFloat(yArr[i]);
           out.writeFloat(cArr[i]);  
         }

       } else {
         out.writeByte(0);      // bb is false

         for (int i=0; i<noOfPoints; i++) {
            out.writeFloat(xArr[i]);
            out.writeFloat(yArr[i]);
            out.writeFloat(cArr[i]);
         }
       }

       // send number of connection points in given time
       out.writeInt(connection.length);

       for (int i=0; i<connection.length; i++) 
         out.writeInt(connection[i]);

    } catch (Exception e) {
       System.out.println("Data sending error(TimeDataCon).");
       //e.printStackTrace();
    } 
  }


  /**
   * update data 
   * @param data TimeDataCon to be appended or updated
   */
  public void updateData(TimeDataCon data, boolean parallel) {
     System.out.println("Update time data function is not implemented.");
  }

  /**
   * return colors
   */
  public float[][] getColors() {
     return colors;
  }

  /**
   * return connection list
   */
  public int[] getConnectionList() {
     int sz = connection.length;
     int[] ret = new int[sz];
     System.arraycopy(connection, 0, ret, 0, sz);

     return ret;
  }

  /**
   * calculate colors
   */
  public void calcColors() {
     if (initialized) return;

     Vector polyvec = new Vector();
     int[] xpts, ypts;
     int index = 0;
     float xfactor = 1.f, yfactor = 1.f; 

     int[] xxArr = new int[xArr.length];
     int[] yyArr = new int[yArr.length];

     colors = new float[321][321];
     for (int i=0; i<321; i++) 
        for (int j=0; j<321; j++) 
           colors[i][j] = Float.NaN;

     if (xmax > xmin)
        xfactor = 320f / (xmax - xmin);
     else 
        return;

     if (ymax > ymin)
        yfactor = 320f / (ymax - ymin);
     else 
        return;

     //////////////////////////////////////////////////
     // STEP 1: normalize to fit 320 x 320 window
     //////////////////////////////////////////////////
     for (int i=0; i<xArr.length; i++) {
        xxArr[i] = (int)((xArr[i] - xmin) * xfactor);
        yyArr[i] = 320 - (int)((yArr[i] - ymin) * yfactor);
     }

     //////////////////////////////////////////////////
     // STEP 2: make the polygon
     //////////////////////////////////////////////////
     xpts = new int[connection.length];
     ypts = new int[connection.length];

     int dummy;
     for (int k=0; k<connection.length; k++) {
        if (connection[k] < 0) {
          polyvec.addElement(new Polygon(xpts, ypts, index));
          index = 0;
        } else {
          xpts[index] = xxArr[connection[k]-1];
          ypts[index] = yyArr[connection[k]-1];
          colors[xpts[index]][ypts[index]] = cArr[connection[k]-1];
          index++; 
        }
     }

     Polygon[] polyArr = new Polygon[polyvec.size()];
     Rectangle[] polybounds = new Rectangle[polyvec.size()];
     polyvec.copyInto(polyArr);
     polyvec = null;

     for (int j=0; j<polyArr.length; j++) {
        int boundsMinX = Integer.MAX_VALUE;
        int boundsMinY = Integer.MAX_VALUE;
        int boundsMaxX = Integer.MIN_VALUE;
        int boundsMaxY = Integer.MIN_VALUE;

        for (int i=0; i < polyArr[j].npoints; i++) {
           int x = polyArr[j].xpoints[i];
           boundsMinX = Math.min(boundsMinX, x);
           boundsMaxX = Math.max(boundsMaxX, x);
           int y = polyArr[j].ypoints[i];
           boundsMinY = Math.min(boundsMinY, y);
           boundsMaxY = Math.max(boundsMaxY, y);
        }

        polybounds[j] = new Rectangle(boundsMinX, boundsMinY,
                    boundsMaxX - boundsMinX, boundsMaxY - boundsMinY);

        //////////////////////////////////////////////////
        // STEP 3: make edge line array
        //////////////////////////////////////////////////
        int dum = polyArr[j].npoints;
        EdgeLine[] lines = new EdgeLine[dum];
        for (int i=1; i< dum; i++) {
           lines[i-1] =
               new EdgeLine(polyArr[j].xpoints[i-1], polyArr[j].ypoints[i-1],
                            polyArr[j].xpoints[i], polyArr[j].ypoints[i],
                   colors[polyArr[j].xpoints[i-1]][polyArr[j].ypoints[i-1]],
                   colors[polyArr[j].xpoints[i]][polyArr[j].ypoints[i]]);

        }
        lines[dum-1] =
            new EdgeLine(polyArr[j].xpoints[dum-1], 
                         polyArr[j].ypoints[dum-1],
                         polyArr[j].xpoints[0], polyArr[j].ypoints[0],
              colors[polyArr[j].xpoints[dum-1]][polyArr[j].ypoints[dum-1]],
              colors[polyArr[j].xpoints[0]][polyArr[j].ypoints[0]]);

//System.out.println("*******> polygon number (lines): " + j + " ( " +
//                    lines.length + " ) " );

        //////////////////////////////////////////////////
        // STEP 4: find intersected edges
        // for a time being, assume convex only
        //////////////////////////////////////////////////
        
        int tmpy1 = polybounds[j].y+1;
        int tmpy2 = polybounds[j].y + polybounds[j].height;

        // scan from top to bottom
        for (int y=tmpy1; y<tmpy2; y++) {

           int intersections = 0;
           float[] secxArr = new float[lines.length];
           int[] lineposArr = new int[lines.length];

           for (int i=0; i<lines.length; i++) {
              lineposArr[i] = -1;
              secxArr[i] = Float.MAX_VALUE; 

              if (lines[i].Intersect(y)) {
                 secxArr[intersections] = lines[i].secx;
                 lineposArr[intersections] = i;
                 intersections++;
              }
	   }

           if (intersections < 2) {
	      System.out.println("Scan line ignored at " + y);
              continue;
           }

           //////////////////////////////////////////////////
           // STEP 5: calc color values along the scanline
           //////////////////////////////////////////////////
           boolean inside = true;

           // bubble sort accroding to x values
           for (int i = secxArr.length; --i>=0; ) {
	     for (int k = 0; k<i; k++) {
		if (secxArr[k] > secxArr[k+1]) {
		    float T = secxArr[k];
		    secxArr[k] = secxArr[k+1];
		    secxArr[k+1] = T;

		    int T2 = lineposArr[k];
		    lineposArr[k] = lineposArr[k+1];
		    lineposArr[k+1] = T2;
		}
             }
	   }

           int prev = (int)lines[lineposArr[0]].secx;
           for (int i=1; i<intersections; i++) {

              EdgeLine line1 = lines[lineposArr[i-1]];
              EdgeLine line2 = lines[lineposArr[i]];

              if (i == 1) {
                if ((line2.secx - line1.secx) < 1.f) 
                  continue;
              }

              if (inside) {
                 for (float x=prev; x<secxArr[i]; x+=1.0f) {

                    float I4 = (float)(y-line1.y2) * line1.c1 / 
                                  (float)(line1.y1 - line1.y2) 
                                  + (float)(line1.y1 - y) * line1.c2 / 
                                  (float)(line1.y1 - line1.y2);

                    float I5 = (float)(y- line2.y2) * line2.c1 / 
                                  (float)(line2.y1 - line2.y2)
                                  + (float)(line2.y1 - y) * line2.c2 / 
                                  (float)(line2.y1 - line2.y2);

                    colors[(int)x][y] = 
                        (x - line1.secx) * I5 /(line2.secx - line1.secx) +
                        (line2.secx - x) * I4 / (line2.secx - line1.secx);

                 }
                 inside = false;

              } else 
                 inside = true;
             
              prev = (int)secxArr[i];             
           }
        } 
        initialized = true;
     }

     //////////////////////////////////////////////////
     // Final STEP: clear up memory
     //////////////////////////////////////////////////
     polyArr = null;
     polybounds = null;
     xpts = ypts = null;
     xxArr = yyArr = null;
  }


  class EdgeLine {
     int x1, x2, y1, y2;
     float c1, c2;            // color values at end vertex
     boolean intersected;

     float secx;
     float cval;


     public EdgeLine(int x1, int y1, int x2, int y2, float c1, float c2) {
        this.x1 = x1;
        this.y1 = y1;
        this.x2 = x2;
        this.y2 = y2;

        this.c1 = c1;
        this.c2 = c2;
     }


     public boolean Intersect(int y) {
        intersected = false;

        if (y1 > y2) 
           if (y >= y2 && y <= y1)
              intersected = true;

        if (y1 < y2) 
           if (y <= y2 && y >= y1)
              intersected = true;

        if (intersected) {
           if (x1 == x2) {
              secx = (float)x1;
           } else {
              float m = (float)(y2-y1) / (float)(x2-x1);
              float b = y1 - m * x1;

              secx = ((float)y - b) / m;
           }
        }

        return intersected;
     } 

  } // end of class EdgeLine

} // end of class

