package sv.ui;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;

/** 
 * LatchButton.java
 * Define latch button
 * @version 1.0     10/15/97
 * @author Byeongseob Ki
 */

public class LatchButton extends Component {

  ActionListener actionListener;      // Post action events to listeners
  String label;                       // The Button's text
  Image btnImg;                       // The Button's image
  
  protected Font font;                // font for string
  protected boolean pressed = false;  // true if the button is detented.
  protected boolean entered = false;  // true if the mouse is on this button.
  protected boolean disabled = false; // true if the button is disabled.

  protected boolean changable = false; // only for changable button
  protected boolean changed = false;

  Image offscreen;                    // double-buffered image
  Image disabledImg;                  // disabled image
  Image changingImg;                  // changing image

  int wid = 50;
  int hgt = 50;

  /**
   * Constructs a LatchButton with no label.
   */
  public LatchButton() {
      this("", null);
  }

  /**
   * Constructs a LatchButton with the specified label.
   * @param label the label of the button
   * @param img the image od the button
   */
  public LatchButton(String label, Image img) {
      this.label = label;
      btnImg = img;
      font = new Font("Helvetica", Font.BOLD, 10);

      // prepare gray image
      makeGrayImage();

      enableEvents(AWTEvent.MOUSE_EVENT_MASK);
  }

  /**
   * gets the label
   * @see setLabel
   */
  public String getLabel() {
      return label;
  }
  
  /**
   * sets the label
   * @see getLabel
   */
  public void setLabel(String label) {
      this.label = label;
      invalidate();
      repaint();
  }

  /**
   * add changing image
   */
  public void addChangingImage(Image img) {
     changable = true;
     changingImg = img;
  }

  /**
   * image change
   */
  public void changeImage(boolean b) {
     changed = b;
     invalidate();
     repaint();
  }

  
  /**
   * null out the offscreen buffer as part of invalidation
   */
  public void invalidate() {
      super.invalidate();
      offscreen = null;
  }

  /**
   * override update to *not* erase the background before painting
   */
  public void update(Graphics g) {
      paint(g);
  }

  /**
   * paints the LatchButton
   */
  public void paint(Graphics g) {
      int w = getSize().width - 1;
      int h = getSize().height - 1;
      FontMetrics fm = getFontMetrics(font);  
   
      if(offscreen == null) {
         offscreen = createImage(getSize().width, getSize().height);
      }
      Graphics og = offscreen.getGraphics();
      og.setClip(0, 0, getSize().width, getSize().height);

      og.setFont(font);
      og.setColor(getBackground());
      if (entered) 
         og.fill3DRect(0, 0, w, h, !pressed);
      else 
         og.fillRect(0, 0, w, h);

      if (btnImg == null) {
         // draw the label centered in the button
         og.setColor(getForeground());
         og.drawString(label,
		       w/2 - fm.stringWidth(label)/2,
		       h/2 + fm.getMaxDescent());
      } else {
         // draw the button image centered in the button
         if (changable && changed && !disabled) {
            if (entered) {
               og.drawImage(changingImg, 
                                   w/2 - btnImg.getWidth(this)/2, 5, this);
               og.setColor(Color.blue.brighter());

            } else {
               og.drawImage(changingImg, 
                                   w/2 - btnImg.getWidth(this)/2, 5, this);
               og.setColor(getForeground());
            }

         } else {
            if (disabled) {
               if (disabledImg != null)
                  og.drawImage(disabledImg, 
                               w/2 - disabledImg.getWidth(this)/2, 5, this);
               og.setColor(getBackground().darker());

            } else if (entered) {
               og.drawImage(btnImg, w/2 - btnImg.getWidth(this)/2, 5, this);
               og.setColor(Color.blue.brighter());

            } else {
               og.drawImage(btnImg, w/2 - btnImg.getWidth(this)/2, 5, this);
               og.setColor(getForeground());
            }
         }

         // draw the label centered in the button
         og.drawString(label, 
                       w/2 - fm.stringWidth(label)/2,
		       btnImg.getHeight(this) + fm.getMaxDescent() + 15);
      }
   
      g.drawImage(offscreen, 0, 0, null);
      og.dispose(); 
  }
  
  /**
   * The preferred size of the button. 
   */
  public Dimension getPreferredSize() {
      int w, h;

      if (btnImg == null) {          // if button image is null

         FontMetrics fm = getFontMetrics(font);
         w = fm.stringWidth(label)+20;
         h = fm.getHeight()+20;

      } else {

         int imgw = btnImg.getWidth(this);
         int imgh = btnImg.getHeight(this);
         
         FontMetrics fm = getFontMetrics(font);
	 w = Math.max(fm.stringWidth(label)+20, imgw);
         h = fm.getHeight() + imgh + 20;
      }
      return new Dimension(w, h);
  }
  
  /**
   * The minimum size of the button. 
   */
  public Dimension getMinimumSize() {
      return new Dimension(wid, hgt);
  }

  /**
   * overided function to handle enabling the button
   * @param b If true, this is enabled; otherwise this button is disabled. 
   */
  public void setEnabled(boolean b) {
     disabled = !b;
     entered = false;
     repaint();
     return;
  }

  /**
   * Adds the specified action listener to receive action events
   * from this button.
   * @param listener the action listener
   */
   public void addActionListener(ActionListener listener) {
       actionListener = AWTEventMulticaster.add(actionListener, listener);
       enableEvents(AWTEvent.MOUSE_EVENT_MASK);
   }
 
   /**
    * Removes the specified action listener so it no longer receives
    * action events from this button.
    * @param listener the action listener
    */
   public void removeActionListener(ActionListener listener) {
       actionListener = AWTEventMulticaster.remove(actionListener, listener);
   }

   /**
    * Paints the button and distribute an action event to all listeners.
    */
   public void processMouseEvent(MouseEvent e) {
       if (disabled) return;

       switch(e.getID()) {
          case MouseEvent.MOUSE_PRESSED:

	    pressed = true;
	    repaint(); 
	    break;

          case MouseEvent.MOUSE_RELEASED:
	    if(actionListener != null) {
	       actionListener.actionPerformed(new ActionEvent(
		   this, ActionEvent.ACTION_PERFORMED, label));
	    }

	    if(pressed == true) {
		pressed = false;
		repaint();
	    }
	    break;

          case MouseEvent.MOUSE_ENTERED:
            if(actionListener != null) {
	       actionListener.actionPerformed(new ActionEvent(
		   this, e.getID(), label));
	    }

            entered = true;
            repaint();
	    break;

          case MouseEvent.MOUSE_EXITED:
	    if(pressed == true) 
		// Cancel! Don't send action event.
		pressed = false;

            entered = false;
            repaint();
	    break;
       }
       super.processMouseEvent(e);
   }


  // Creates a gray image (displayed when the button is disabled).
  private void makeGrayImage() {
     if (btnImg == null)
	return;

     ImageFilter filter = new GrayFilter();
     ImageProducer producer = 
           new FilteredImageSource(btnImg.getSource(), filter);
     disabledImg = createImage(producer);
  }

 
} // end of class



/**
 * class GrayFilter is used to create gray scaled images.
 */
class GrayFilter extends RGBImageFilter {
   public GrayFilter() { 
      canFilterIndexColorModel = true; 
   }

   public int filterRGB(int x, int y, int rgb) {
      int c;
      int a = rgb & 0xff000000;
      int r = (rgb & 0xff0000) >> 16;
      int g = (rgb & 0x00ff00) >> 8;
      int b = (rgb & 0x0000ff);

      if (r == g && g == b)
	c = r;
      else
	c = (r * 299)/100 +  (g * 587)/100 + (b * 114)/100;

      return a | (c << 16) | (c << 8) | c;
   }
}
        
