package sv.util;


/** 
 * CubicSpline.java
 * cubic spline algorithm
 * @version 1.0
 * @author Byeongseob Ki
 */

public class CubicSpline {
  private int noOfKnots;      // number of knots
  private float xval[];      // x values
  private float fval[];      // f(x) values

  private float Avec[];         // coefficient matrix A
  private float Bvec[];         // coefficient matrix B   
  private float Cvec[];         // coefficient matrix C
  private float h[];
  
  /**
   * constructor
   * @param xval x values
   * @param fval f(x) values
   */
  public CubicSpline(float xval[], float fval[]) {
    noOfKnots = xval.length;
    this.xval = xval;
    this.fval = fval;

    Avec = new float[noOfKnots];
    Bvec = new float[noOfKnots];
    Cvec = new float[noOfKnots];
    h = new float[noOfKnots];

    float[] a = new float[noOfKnots];
    float[] b = new float[noOfKnots];
    float[] c = new float[noOfKnots];
    float[] dd = new float[noOfKnots];
    float[] d = new float[noOfKnots];
    float m;

    // form diffrences    
    for (int i=0; i<noOfKnots-1; i++) {
       h[i] = xval[i+1] - xval[i];
       dd[i] = (fval[i+1] - fval[i])/h[i];
    }
    for (int i=1; i<noOfKnots-1; i++)
       d[i] = 6.0f * (dd[i] - dd[i-1]);
  
    // Obtain coeffiecient matrix
    for (int i=1; i<noOfKnots-1; i++) {
       b[i] = c[i] = h[i];
       a[i] = 2.0f * (h[i-1] + h[i]);
    }

    // Solve tridiagonal system
    for (int i=1; i<noOfKnots-2; i++) {
       m = c[i]/a[i];
       c[i] = 0.0f;

       a[i+1] = a[i+1] - m*b[i];
       d[i+1] = d[i+1] - m*d[i];
    }
    Avec[noOfKnots-2] = d[noOfKnots-2] / a[noOfKnots-2];
    for (int i=noOfKnots-3; i>=1; i--) 
      Avec[i] = (d[i] - b[i]*Avec[i+1]) / a[i];

    Avec[0] = Avec[noOfKnots-1] = 0.0f;

    // Compute spline coefficients
    for(int i=0; i<noOfKnots-1; i++) {
       Bvec[i] = fval[i]/h[i] - Avec[i]*h[i]/6.0f;
       Cvec[i] = fval[i+1]/h[i] - Avec[i+1]*h[i]/6.0f;
    }

//for(int i=0; i<noOfKnots; i++) 
//System.out.println(i + ": " + Avec[i] + "   " + Bvec[i] + "    " + Cvec[i]);
  }

  /**
   * evaluation of cubic spline
   * @param val x value to compute s(x)
   */
  public float evaluate(float val) {
    int k = 0;
    float u0, u1;

    for (int i=0; i<noOfKnots; i++) {
      if (val >= xval[i] && val <= xval[i+1]) {
         k = i;
         break;
      }
    }

    u0 = val - xval[k];
    u1 = val - xval[k+1];
    return ((Avec[k+1] * u0*u0*u0 - Avec[k] * u1*u1*u1) / (6.0f*h[k]) - Bvec[k]*u1 + Cvec[k]*u0);

  }
}





