import java.awt.*;
import java.awt.event.*;
import java.net.*;
import java.io.*;
import java.util.*;

import com.sun.java.swing.*;
import com.sun.java.swing.event.*;
import com.sun.java.swing.border.*;

import sv.ui.*;
import sv.drawer.*;
import sv.kernel.*;

/** 
 * svserver.java
 * scientific data visualization server
 * @version 1.0   10/15/97 
 * @author Byeongseob Ki
 */

class svserver extends Frame 
                        implements Runnable, ActionListener {
  Thread runner;

  static svserver instance;

  private static final int DEFAULT_PORT = 7654;
  private ServerSocket mainSock = null;
  private static Vector conVec = new Vector(5, 2); // connection vector

  /**
   * server port
   */
  public int this_port = DEFAULT_PORT;

  /**
   * Configuration handler (this is not static!!!)
   */
  public Configuration configHandler;

  /**
   * frame management handler 
   */
  public static FrameManager fmanager;

  /**
   * collaboration management
   */
  public static CollabManager cmanager;

  /**
   * Merge filter handler
   */
  public static MergeFilterHandler mergeHandler;

  /**
   * Filter menu handler
   */
  public static FilterMenuHandler menuHandler;

  /**
   * Clipboard handler
   */
  public static ClipboardHandler clipboard;

  /**
   * Drawing tool and properties
   */
  public DrawerUI dtool;
  public ShapeProperties shapes;


  /** 
   * Virtual desktop handler
   */
  public DesktopControl controller;

  /** 
   * Main system buttons
   */
  public MainButtons mainb; 
  
  // private objects
  SimpleMonitor monitor;     // for message output
  JTabbedPane tabbedPane;    // tabbed panel

  /**
   * svserver class constructor
   * @param port server port
   */
  public svserver(int port, String setting) {
     super("NPAC Scivis Server");

     configHandler = new Configuration(setting);

     this_port = port;
     try {
        mainSock = new ServerSocket(port);
     } catch(Exception e) { 
        System.err.println("\nCan't create system socket."); 
        System.err.println("You'd better choose another port and restart Scivis");
     }  

     ///////////////////////////////////////////////////////////////////
     // make gui
     BorderLayout bl = new BorderLayout();
     bl.setVgap(3);
     setLayout(bl);

     // Create a tab pane
     tabbedPane = new JTabbedPane();
     tabbedPane.setFont(new Font("Helvetica", Font.BOLD, 12));

     tabbedPane.addTab("System Menus", mainb = new MainButtons(this));
     tabbedPane.setSelectedIndex(0);
     tabbedPane.addTab("Desktop Control", 
                      controller = new DesktopControl(this));
     add("Center", tabbedPane);

     // Add Tab change listener
     createTabListener();

     add("South", monitor = new SimpleMonitor());

     setResizable(false);
     //////////////////////////////////////////////////////////////////////

     fmanager = new FrameManager(this);
     cmanager = new CollabManager(this);
     clipboard = new ClipboardHandler();
     mergeHandler = new MergeFilterHandler(this);
     menuHandler = new FilterMenuHandler(configHandler.getMenuFile());
     shapes = new ShapeProperties();

     // shared instance
     instance = this;
    
     setIconImage(SVUIC.iconImg);    

     // print out system status
     message("Server is ready to port " + port, false);
  }


  /**
   * display the instruction and message
   * @param msg message
   * @param critical if true, message will be displayed with red color
   */
  public void message(String msg, boolean critical) {
    monitor.display(msg, critical);
  }

  void createTabListener() {
	// add listener to know when we've been shown
        ChangeListener changeListener = new ChangeListener() {
            public void stateChanged(ChangeEvent e) {
                JTabbedPane tab = (JTabbedPane) e.getSource();
                int index = tab.getSelectedIndex();
                Component currentPage = tab.getComponentAt(index);

                if(currentPage == mainb) {
		  message("Scivis system tools and utilities.", false);
		} else if(currentPage == controller) {
                  message("Scivis desktop control.", false);
		} 
            }
        };
        tabbedPane.addChangeListener(changeListener);
  }


  /**
   * return this object
   */
  public static svserver sharedInstance() {
    return instance;
  }

  /**
   * load image icon
   */
  public ImageIcon loadImageIcon(URL filename, String description) {
    return new ImageIcon(filename, description);
  }


  /** 
   * do work when icon was clicked
   * @param e ActionEvent
   */
  public void actionPerformed(ActionEvent e) {

     String command = e.getActionCommand();

     if (e.getID() == MouseEvent.MOUSE_ENTERED) {
        if (command.equals("Connect")) { 
           message("Connect to other server for collaboration.", false);
        } else if (command.equals("Open")) {
           message("Read data from a local file.", false);
        } else if (command.equals("Drawing")) {
           message("Invoke drawing tool.", false);
        } else if (command.equals("Editor")) {
           message("Invoke filter editor.", false);
        } else if (command.equals("Guide")) {
           message("Help and guide of this system.", false);
        } else if (command.equals("Kill All")) {
           message("Kill all visualization windows.", false);
        } else if (command.equals("Check All")) {
           message("Toggle to send command to all windows.", false);
        } else if (command.equals("Exit")) {
           message("Exit the system.", false);
        } 

     } else if (e.getID() == ActionEvent.ACTION_PERFORMED) {
        if (command.equals("Connect")) { 
           cmanager.clickedConnectButton();

        } else if (command.equals("Open")) {

           FileDialog openFileDialog = new FileDialog(this,
                         "Open local data file", FileDialog.LOAD);
           openFileDialog.show();

           String fname = openFileDialog.getFile();
           if (fname == null) 
              return;

           fname = openFileDialog.getDirectory() + fname;
           
           DataInputStream in = null;
           try {
             in = new DataInputStream(
                  new BufferedInputStream(new FileInputStream(fname)));
           } catch (Exception E) {
               System.out.println("Can't find file " + fname + " ...");
               return;
           }
           SVFileReader.readFromInput(this, in, 1);
         
        } else if (command.equals("Drawing")) {
           dtool = new DrawerUI(this);
           dtool.pack();
           dtool.show();
           fmanager.setDrawingMode(true);
           mainb.drawBtn.setEnabled(false);

        } else if (command.equals("Editor")) {
           FilterEditor pad = new FilterEditor(this);
           pad.pack();
           pad.show();
           mainb.editBtn.setEnabled(false);

        } else if (command.equals("Guide")) {
           HelpFrame f = new HelpFrame(this);
           f.setSize(560,600);
           f.show();
           f.setInitialURL(getClass().getResource("helps/index.html"));
           mainb.helpBtn.setEnabled(false);
        
        } else if (command.equals("Kill All")) {
           fmanager.killAllFrames(true);

        } else if (command.equals("Check All")) {
           fmanager.clickedGlobalCheckButton(true);

        } else if (command.equals("Exit")) {
           cmanager.clickedExitButton();

           //clean up the vector 
           for (int i=0; i<conVec.size(); i++) {
              ServerManager manager = (ServerManager)conVec.elementAt(i);
              manager.stop();
              manager = null;
           }
 
           this.dispose();
           System.exit(1);
        } 
     } 
  }

  /**
   * waits for a connection then spawns a ServerManager to deal with it
   */
  public void run() {

     // continuously listen the socket
     while (true) {
        try {
           Socket this_connection = mainSock.accept();
	
           ServerManager manager = new ServerManager(this_connection, this);
           manager.start();
           conVec.addElement(manager);

           //clean up the vector if needed
           for (int i=0; i<conVec.size(); i++)
              if (!((ServerManager)(conVec.elementAt(i))).isAlive())
                  conVec.removeElementAt(i);
        } catch(Exception e) {
           System.err.println("Can't accept any connection from the client.");
           return;
        }
     }
  }


  /**
   * start thread
   */   
  public void start() {
    if (runner == null); {
      runner = new Thread(this);
      runner.start();
    }
  }

  /**
   * stop thread
   */
  public void stop() {
    if (runner != null) {
      runner.stop();
      runner = null;
    }
  }

  // parse the port argument and start server
  public static void main(String[] args) {
     int serverPort = DEFAULT_PORT;
     String setting = null;
     int i = 0;

     try {

        String vers = System.getProperty("java.version");
        if (vers.compareTo("1.1.2") < 0) {
            System.out.println("!!!WARNING: Scivis must be run with a " +
                               "1.1.2 or higher version VM!!!");
        }

        while (i < args.length) {

         if (args[i].startsWith("-")) {
            String arg = args[i++];
            if (arg.equals("-s")) 
               setting = args[i++];
            else if (arg.equals("-p")) 
               serverPort = Integer.valueOf(args[i++]).intValue();
             
          }  
        }

     } catch(Exception e) {
        serverPort = DEFAULT_PORT;
     }

     if (setting == null) 
       System.out.println("\nNo configuration files are declared. Using defaults...");

     svserver server = new svserver(serverPort, setting);
     server.setSize(320, 280);
     server.show();  
     server.start();

     server.validate();
     server.repaint(); 
  }


  /////////////////////////////////////////////////////////
  // To shutdown when run as an application.  This is a
  // fairly lame implementation.   A more self-respecting
  // implementation would at least check to see if a save
  // was needed.
  // **************** FOR DATAPAD ********************
  /////////////////////////////////////////////////////////
   protected static final class AppCloser extends WindowAdapter {
      public void windowClosing(WindowEvent e) {
      }
   }

} // end of class
